/*  regrid.c -- Regrid a raster image containing GIS-type data

NAME:
    REGRID

PURPOSE:
    REGRID uses a standard LAS mapping grid file to aggregate data such
    as land use category or surface properties into grid cells of a
    different size or in a different map projection.  The program
    supports three modes of aggregation:  finding the mode (most
    frequently occuring) of the values for all the input pixels which
    contribute to a given output pixel, finding the mean value over all
    the input pixels, or finding the fraction of input pixels which are
    in each of a small number of different classes.

EXTERNAL ASSOCIATIONS:
    REGRID uses a standard LAS mapping grid to determine which pixels of
    the input image contribute to each pixel of the output image.

SUBCOMMANDS:
    MEAN	Assign to each output pixel the mean of the values for
		all input pixels contained within the output pixel.
    MODE	Assign to each output pixel the value which occurs most
		frequently in the input pixels contained within the
		output pixel.  The input image may be either a
		single-band image containing, in each pixel, the class
		(category) number for that pixel, or a multi-band image
		similar to that produced using aggregatiion function
		FRACT.
    FRACT	Create a multi-band output image for which band n gives,
		for each output pixel, the fraction of input pixels
		which have value n.

ARGUMENTS:
    INFILE	Name of input image file.
    GRID	Name of mapping grid file.
    OUTFILE	Name of output file.
    WINDOPT	Image to use for defining output window:  IN or OUT.
    OWINDOW	Output window specification if WINDOPT = OUT, in order
		SL, SS, NL, NS.
    ODTYPE	Output data type:  SAME, BYTE, I*2, I*4, or R*4.
    FILL	Fill value for output pixels whose mapped locations
		fall enirely outside the input image.
    NODATA	Value of input pixel which is to be omitted from mean.
    WCLASS	Input class(es) to be assigned weights other than 1.0.
    WEIGHT	Weight(s) to be assigned class(es) specified by WCLASS.
    TCLASS	Input class(es) for which threshold value(s) are to be
		used when determining mode.
    THRESH	Threshold value(s) to use with class(es) specified by
		TCLASS.
    REALLOC	Name of file containing class reallocation
		specifications and/or weight and/or threshold values.

FILE REFERENCES:
    INFILE	File containing input image.
    GRID	File containing mapping grid.
    OUTFILE	File to receive regridded output.
    REALLOC	File containing reallocation, weighting, and
		thresholding arguments.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    1994-07	Added support for class reallocation, R. A. White,
		PSU/ESSC.
    1999-02	Added subcommands to specify aggregation mode, raw.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Sun Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in regrid.pdf.

EXTERNAL FUNCTIONS CALLED:

(Program Specific Functions)
    rgd_aggreg	Aggregate data using specified function.
    rgd_bldddr	Build DDR for output image.
    rgd_ckddr	Get input DDR and check that data type is compatible
		with requested aggregation mode.
    rgd_ckmnmx	Update minimum/maximum coordinate values (via
		rgd_outbox, rgd_inrect).
    rgd_getarg	Ingest program arguments.
    rgd_getwtr	Get class weighting, thresholding, and reallocation
		specifications.
    rgd_incorn	Get input coords of corners of each output pixel on line
		separating two rows of output pixels (via rgd_aggreg).
    rgd_inrect	Determine input rectanagle for each output line.
    rgd_outbox	Determine output window corresponding to input window
		(via rgd_setgrd).
    rgd_pxarea	Compute overlap area of input and output pixel (via
		rgd_tally).
    rgd_setgrd  Set up mapping grid and compute gradients.
    rdg_tally	Tally contributions of each input pixel to region
		covered by output pixel and compute output value (via
		rgd_aggreg).
    rgd_trmgrd	Trim mapping grid to specified output window (via
		rgd_setgrd).

    Calling Hierarchy for Program Specific Functions:

	regrid
	    rgd_getarg
		rgd_getwtr
	    rgd_ckddr
	    rgd_setgrd
  		rgd_trmgrd
		rgd_outbox
		    rgd_ckmnmx
	    rgd_inrect
		rgd_ckmnmx
	    rgd_aggreg
		rgd_incorn
		rgd_tally
		    rgd_pxarea
	    rgd_bldddr

ALGORITHM DESCRIPTION:
    Initialize las/tae interface <c_inlas>
    Ingest program agruments <rgd_getarg>
    Get input image DDR to get number of classes and output bands
      <rgd_ckddr>
    Set up mapping grid and compute gradients <rgd_setgrd>
    Determine input rectanagle for each output line <rgd_inrect>
    Aggregate data using specified function <rgd_aggreg>
    Create output DDR from mapping grid data and output window
      <rgd_bldddr>
    Update history record <c_hstout>
    Report processing completion <c_complt>
	
    									
***************   End of Prolog for REGRID   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"geompak.h"	/* Geom. transform package definitions	*/
#include	"util.h"	/* LAS utiltiy function definitions	*/
#include	"regrid.h"	/* Definitions unique to REGRID		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    void rgd_aggreg (long aggrcode, long bands[], float fill,
		     char ifminmax, struct DDR *inddr, char infile[],
		     long iwindow[], long lastil[], long maxnli,
		     long nbndin, long nbndout, long nclass, long ncol,
		     long nrow, long odcode, float olg[], float osg[],
		     char outfile[], long owindow[], long  pmsgtim,
		     long startil[], struct VTEX vertex[], long windopt,
		     struct WTR *wtr, double **maxval, double **minval);
    void rgd_bldddr (struct GEOGRID *grid, char ifminmax,
		     double maxval[], double minval[], long nbndout,
		     char outfile[], long owindow[]);
    void rgd_ckddr  (long aggrcode, char *infile, long bands[],
		     long nbands, struct DDR *inddr, long *nclass,
		     long *nbndout);
    void rgd_getarg (struct PARBLK *vblock, long *aggrcode,
		     char *infile, long bands[], long *nbands,
		     long iwindow[4], char *gridfile, char *outfile,
		     long *windopt, long owindow[4], long *odcode,
		     float *fill, struct WTR *wtr, char *ifminmax,
		     long *pmsgtim);
    void rgd_inrect (long nlo, long nrow, long ncol, float olg[],
		     float osg[], struct VTEX v[], long windopt,
		     long iwindow[4], long *maxnli, long **startil,
		     long **lastil);
    void rgd_setgrd (char *gridfile, long windopt, long iwindow[4],
		     long owindow[4], struct GEOGRID *grid, long *nrow,
		     long *ncol, float **outline, float **outsamp,
		     struct VTEX **vertex);
									/*
***************   Main Routine for REGRID   ***************	*/

int main ()
    {
	                                              			/*
Local Variable Definitions:						*/

    long	aggrcode;	/* Aggregation function code */
    long	bands[MAXBND+1]; /* Array of input band numbers */
    float	fill;		/* Fill value for pixels outside input */
    struct GEOGRID
		grid;		/* Mapping grid */
    char	gridfile[CMLEN]; /* Name of input mapping grid file */
    char	ifminmax;	/* Set TRUE if $MINMAX = "yes" */
    struct DDR	inddr;		/* DDR for input image */
    char	infile[CMLEN];	/* Host name of input image file */
    long	iwindow[4];	/* Input window:  sl, ss, nl, ns */
    long	*lastil;	/* Array of last input line for each
				   output line */
    long	maxnli;		/* Maximum input lines for any output */
    double	*maxval;	/* Array of maximum pixel values for
				   each output band */
    double	*minval;	/* Array of minimum output values */
    long	nbands;		/* Number of input bands */
    long	nbndout;	/* Number of bands in output image */
    long	nclass;		/* Number of classes */
    long	ncol;		/* Number of columns in mapping grid */
    long	nrow;		/* Number of rows in mapping grid */
    long	odcode;		/* LAS data type code of output image */
    long	one = 1;	/* Value 1 */
    float	*outline;	/* Output line number on each grid row */
    float	*outsamp;	/* Output sample number on each grid
				   column */
    char	outfile[CMLEN];	/* Host name of output image file */
    long	owindow[4];	/* Output image window */
    long 	pmsgtim;	/* Seconds between processing messages */
    long	stdlun;		/* Unit number for standard output */
    long	*startil;	/* Array of first input line for each
				   output line */
    struct                
      PARBLK	vblock;		/* TAE parameter block */
    struct VTEX *vertex;	/* Input coordinates and derivatives at
				   each vertex of mapping grid */
    long	windopt;	/* Code for windowing option */
    struct WTR  wtr;		/* Weighting, thresholding, and class
				   reallocation specifications */

/*  Initialize TAE and LAS  */                                          

    c_inlas (&vblock, &stdlun);

/*  Ingest program arguments  */

    rgd_getarg (&vblock, &aggrcode, infile, bands, &nbands, iwindow,
      gridfile, outfile, &windopt, owindow, &odcode, &fill, &wtr,
      &ifminmax, &pmsgtim);

/*  Get input image DDR and check data type against aggregation mode */

    rgd_ckddr (aggrcode, infile, bands, nbands, &inddr, &nclass,
      &nbndout);

/*  Set up geometric mapping grid to match requested window  */

    rgd_setgrd (gridfile, windopt, iwindow, owindow, &grid, &nrow,
      &ncol, &outline, &outsamp, &vertex);

/*  Determine input rectangle for each output line  */

    rgd_inrect (owindow[2], nrow, ncol, outline, outsamp, vertex,
      windopt, iwindow, &maxnli, &startil, &lastil);

/*  Aggregate data using specified function  */

    rgd_aggreg (aggrcode, bands, fill, ifminmax, &inddr, infile,
      iwindow, lastil, maxnli, nbands, nbndout, nclass, ncol, nrow,
      odcode, outline, outsamp, outfile, owindow, pmsgtim, startil,
      vertex, windopt, &wtr, &maxval, &minval);

/*  Build output DDR from mapping grid and window data  */

    rgd_bldddr (&grid, ifminmax, maxval, minval, nbndout, outfile,
      owindow);

/*  Update history record and report processing completion  */

    c_hstout (&vblock, outfile, (void *) infile, &one, " ");
    c_complt ("regrid", (void *) outfile, &one);

    return 0;	/* to supporess compiler warning */

    }			/* End of Program REGRID */
