/*  rgd_aggreg.c -- Aggregate data using specified function

NAME:
    rgd_aggreg ()

PURPOSE:
    Called by program REGRID to aggregate input pixels into each output
    pixel using the aggregation function specified by the user.

ARGUMENTS:
    long    aggrcode	(in)	Aggregation function code.
    long    bands[]	(in)	Array of input band numbers.
    float   fill	(in)	Fill value for pixels outside input.
    char    ifminmax	(in)	Set TRUE if $MINMAX = "yes".
    struct DDR
	    *inddr	(in)	DDR for input image.
    char    infile[]	(in)	Host name of input image file.
    long    iwindow[]	(in)	Input window:  sl, ss, nl, ns.
    long    lastil[]	(in)	Array of last input line for each
				output line.
    long    maxnli	(in)	Maximum input lines for any output.
    long    nbndin	(in)	Number of input bands.
    long    nbndout	(in)	Number of bands in output image.
    long    nclass	(in)	Number of classes.
    long    ncol	(in)	Number of columns in mapping grid.
    long    nrow	(in)	Number of rows in mapping grid.
    long    odcode	(in)	LAS data type code of output image.
    float   olg[]	(in)	Output line number on each grid row.
    float   osg[]	(in)	Output sample number on each grid
				column.
    char    outfile[]	(in)	Host name of output image file.
    long    owindow[]	(in)	Output image window.
    long    pmsgtim	(in)	Seconds between processing messages.
    long    startil[]	(in)	Array of first input line for each
				output line.
    struct VTEX
	    vertex[](in)	Input coordinates and derivatives at
				each vertex of mapping grid.
    long    windopt	(in)	Code for windowing option.
    struct WTR
	    *wtr	(in)	Specifications for weighting,
				thresholding, and reallocating classes.
    double  **maxval	(out)	Array of maximum pixel values for
				each output band.
    double  **minval	(out)	Array of minimum output values.

RETURN VALUE:
    None.

FILE REFERENCES:
    <infile>.IMG	(read)		Input image to be regridded.
    <outfile>.IMG	(write)		Regridded output.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    1994-07	Added class reallocation, R. A. White, PSU/ESSC.
    1996-02	Fixed buffer wrap error, R. A. White, PSU/ESSC.
    2001-08	Added support for NODATA in mode MEAN, raw.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in regrid.pdf.

EXTERNAL FUNCTIONS CALLED:

(Program Specific Functions)
    rgd_incorn	Get input coords of corners of each output pixel on line
		separating two rows of output pixels.
    rgd_tally	Tally contributions of each input pixel to region
		covered by one output pixel and compute output value.

ALGORITHM DESCRIPTION:
    If aggregation function is FRACT
	Determine scaling factor for output pixel type
    Open output file for sequential write <c_eopens>
    Allocate buffer space for output <c_ealloc>
    Define file group for output <c_egroup>
    Allocate space for accumulating class or band totals
    Try to open input for random read of entire input window <c_eopenr>
    If successful
	Try to allocate input buffer for entire input window
    If insufficient memory for entire input window
	Allocate input buffer for maximum window per output line
	Open input for random read of maximum window per output line
	  <c_eopenr>
	If not successful
	    Suggest user process image in strips
	    Abort
    Determine size of input pixels
    If entire input fits in buffer
	Read entire input image <c_eread>
    Get input coords of upper corners of each pixel on first output
      line <rgd_incorn>
    If computation of output minima and maxima requested
	Initiate min/max arrays <c_inmnmx>
    If processing progress messages are wanted
	Initiate messages and start them <c_inittm, c_startm>
    Get buffer offsets for first output line <c_estep>
    For each output line
        If entire input image not in buffer
	    Retrieve first, last input line needed
	    Compute buffer wrap line, if any
	    For each segment (first, wrap) and (wrap +1, last)
		If pair stop < first line in buffer or pair start > last
		  line in buffer
		    Set read block from first to last line of pair
		    Update first, last line in buffer
		Else
		    If pair start < first line in buffer
			Set read block from first line to buf start
			Update first line in buffer
		    If pair stop > last line in buffer
			Set read block from buf end to last line
			Update last line in buffer
	For each input block to be read
	    For each input band
		Read block of input data into buffer
	If read error occurred
	    Inform user
	    Abort
	Get input coords of lower corners of each output pixel on line
	  <rgd_incorn>
	For each output pixel
	    Tally contributions of each input pixel and set output
	      pixel as specified by FUNCTION agrument <rgd_tally>
	Write output line <c_estep>
	If output minima/maxima requested
	    For each output band
		Update maximum and minimum for band <c_pxmnmx>
	Save corners at bottom of this line as top of next line
    Close input file <c_eclose>
    Close output file <c_egclse>
    									
***************   End of Prolog for rgd_aggreg   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String processing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"pixman.h"	/* LAS pixel manipulation functions	*/
#include	"util.h"	/* LAS utility functions		*/
#include	"regrid.h"	/* Definitions unique to REGRID		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    void rgd_incorn (float olcoord, long ncol, float olg[], float osg[],
		     struct VTEX vertex[], float incorn[]);
    void rgd_tally (long aggrcode, long bndndx[], float fill,
		    long idcode, long idsize, unsigned char *inbuf,
		    long maxnli, long nbndin, long nbndout, long nclass,
		    long nli, long nsi, long odcode, long odsize,
		    long offsets[], float ovtex[4][2], float scalefact,
		    struct WTR *wtr, float accum[],
		    unsigned char *outbuf);
									/*
External Variable Declarations:						*/
    extern	long	curent;	/* Current entity for processing
				   progress meassages */
    extern	long	curcat;	/* Current category for processing
				   progress meassages */
									/*
***************   Start of Function rgd_aggreg   ***************	*/

void FUNCTION rgd_aggreg (
    long	aggrcode,	/* Aggregation function code */
    long	bands[],	/* Array of input band numbers */
    float	fill,		/* Fill value for pixels outside input */
    char	ifminmax,	/* Set TRUE if $MINMAX = "yes" */
    struct DDR	*inddr,		/* DDR for input image */
    char	infile[],	/* Host name of input image file */
    long	iwindow[],	/* Input window:  sl, ss, nl, ns */
    long	lastil[],	/* Array of last input line for each
				   output line, relative to window ulc */
    long	maxnli,		/* Maximum input lines for any output */
    long	nbndin,		/* Number of input bands */
    long	nbndout,	/* Number of bands in output image */
    long	nclass,		/* Number of classes */
    long	ncol,		/* Number of columns in mapping grid */
    long	nrow,		/* Number of rows in mapping grid */
    long	odcode,		/* LAS data type code of output image */
    float	olg[],		/* Output line number on each grid row */
    float	osg[],		/* Output sample number on each grid
				   column */
    char	outfile[],	/* Host name of output image file */
    long	owindow[],	/* Output image window */
    long 	pmsgtim,	/* Seconds between processing messages */
    long	startil[],	/* Array of first input line for each
				   output line, relative to window ulc */
    struct VTEX vertex[],	/* Input coordinates, relative to window
				   ulc, and derivatives at each vertex
				   of mapping grid */
    long	windopt,	/* Code for windowing option */
    struct WTR	*wtr,		/* Class weights, thresholds, and
				   reallocations */
    double	**maxval,	/* Array of maximum pixel values for
				   each output band */
    double	**minval)	/* Array of minimum output values */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	access;		/* Access mode for file open */
    float	*accum;		/* Accumulators for area-weighted
				   totals */
    long	badread;	/* Flag set true if input read error */
    long	blklim[3][2];	/* First and last line, relative to top
				   of input window, for each block of
				   input to be read */
    long	bndndx[MAXBND];	/* Index into inbuf for start of each
				   band */
    long	dummy;		/* Dummy argument for call */
    long	eband;		/* Number of current input band */
    unsigned char
		*ebuf;		/* Pointer to location in buffer to
				   receive next input block */
    long	eline;		/* Index of first line in block to read
				   relative to top of input window */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    long	flbuf;		/* First input line in buffer */
    struct GDESC
		*gdesc;		/* Output image group descriptor */
    long	i;		/* Loop counter/index */
    long	idcode;		/* Input image datatype */
    long	idsize;		/* Bytes per input pixel */
    unsigned char
		*inbuf;		/* Input image buffer */
    float	*incorn;	/* Input coordinates (l,s), relative to
				   ulc of input window, of upper or
				   lower corners of each output pixel on
				   line, followed by (l,s) of
				   lower/upper corners */
    struct FDESC
		*infd;		/* File descriptor for input */
    long	inlim[2][2];	/* First and last input line for input
				   needed before and after wrap,
				   relative to top of input windwow */
    long	j;		/* Loop counter/index */
    long	llbuf;		/* Last input line in buffer */
    float	*llcptr;	/* Pointer into incorn for output llc */
    char	msgbuf[ERRLEN];	/* Buffer for building error message */
    long	mtype;		/* Type of processing progress message */
    long	nblkrd;		/* Number of input blocks to read */
    long	nlblk;		/* Number of lines in current read block */
    long	nli;		/* Number of lines in input window */
    long	nlo;		/* Number of output lines */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    long	nsi;		/* Number of samples in input window */
    long	nso;		/* Number of samples per output line */
    unsigned char
		*obufbase;	/* Location in output buffer to receive
				   next pixel for first band */
    long	odsize;		/* Bytes per output pixel */
    long	offsets[MAXBND];/* Offsets into output buffer */
    long	oline;		/* Current output line */
    long	one = 1;	/* Constant one */
    long	opt[2] = {0, 0};/* Options for c_eopens () */
    long	osamp;		/* Current output sample */
    unsigned char
		*outbuf;	/* Output buffer */
    long	outbufsz;	/* Size of output buffer */
    struct FDESC
		*outfd[2];	/* File descriptor for output */
    float	ovtex[4][2];	/* Input line and sample coordinates of
				   vertices of output pixel, going
				   clockwise around input pixel,
				   relative to input window ulc */
    char	*pad = 0;	/* Temporary stack allocation to leave
				   pad after allocating input buffer */
    long	pxsize[NDTYPES+1]; /* Pixel lengths for each pixel type */
    float	scalefact = 0;	/* Scaling factor for FRACT output */
    float	*ulcptr;	/* Pointer into incorn for output ulc */
    long	wrap;		/* Number of times input wraps */
    long	zero = 0;	/* Constant zero */

/*  Determine sizes of input and output pixels  */

    idcode = inddr->dtype;
    pxsize[0] = 0;
    pxsize[EBYTE] = pxsize[ECHAR] = 1;
    pxsize[EWORD] = 2;
    pxsize[ELONG] = pxsize[EREAL] = 4;
    pxsize[EDOUBLE] = 8;
    if (odcode == 0)
	odcode = idcode;
    idsize = pxsize[idcode];
    odsize = pxsize[odcode];

/*  Determine scaling factor for FRACT output  */

    if (aggrcode == RGD_FRACT)
	{
        if ((odcode == ECHAR) || (odcode == EBYTE))
	    scalefact = 100.0;
	else if (odcode == EWORD)
	    scalefact = 10000.0;
	else if (odcode == ELONG)
	    scalefact = 1000000.0;
	else
	    scalefact = 1.0;
	}

/*  Open output file for sequential writes  */

    nlo = owindow[2];
    nso = owindow[3];
    access = IWRITE;
    if (c_eopens (outfd, outfile, &nbndout, &one, &one, &nlo, &nso,
      offsets, &access, &odcode, &zero, opt) != E_SUCC)
	ABORT;
    outfd[1] = 0;

/*  Allocate output buffers and connect to output group  */

    if ((outbuf = (unsigned char *) c_ealloc (outfd, &outbufsz))
      == NULL) 
	ABORT;
    if (c_egroup (outfd, &gdesc, outbuf, &outbufsz) != E_SUCC)
	ABORT;

/*  Allocate non-image temporary arrays  */

    incorn = (float *) malloc (4 * (nso + 1) * sizeof (float));
    *maxval = (double *) malloc (2 * nbndout * sizeof (double));
    *minval = *maxval + nbndout;
    if ( ((accum = (float *) malloc (((wtr->ndflag > 0) ? 2 : 1) *
	     nclass * sizeof (float))) == NULL)
      || ((pad = (char *) malloc (100000)) == NULL) )
	{
	sprintf (msgbuf, "%s%s%ld%s%s%s", "Unable to allocate",
	  "accumulators for", nclass, " class area accumulators",
	  " -- compress data range (e.g., using program MAP) and",
	  " rerun REGRID");
	c_errmsg (msgbuf, "regrid-manyclass", &nonfatal);
	ABORT;
	}

/*  Try to open input image for random read of entire input window  */

    inbuf = NULL;
    dummy = 0;
    nli = iwindow[2];
    nsi = iwindow[3];
    access = IREAD;

    if (c_eopenr (&infd, infile, &dummy, &iwindow[0], &iwindow[1], &nli,
      &nsi, &access, &idcode, &zero, &nli) == E_SUCC)
	{
	if ((inbuf = (unsigned char *)
	  malloc (idsize * nsi * nli * nbndin)) == NULL)
	    c_eclose (&infd);
	}

/*  If unsuccessful, open for max lines needed per output line  */

    if (inbuf != NULL)
	maxnli = nli;
    else
	{
	sprintf (msgbuf, "Entire input window too large for memory %s",
	  "-- will read input incrementally");
	c_errmsg (msgbuf, "informational", &nonfatal);
	if (c_eopenr (&infd, infile, &dummy, &iwindow[0], &iwindow[1],
	  &nli, &nsi, &access, &idcode, &zero, &maxnli) == E_SUCC)
	    {
	    inbuf = (unsigned char *)
	      malloc (idsize * nsi * maxnli * nbndin);
	    }
	if (inbuf == NULL)
	    {
	    c_errmsg ("Cannot allocate input buffer space",
	      "regrid-intoobig", &nonfatal);
	    sprintf (msgbuf, "REGRID needs %ld input lines of %ld",
	      maxnli, nsi * idsize * nbndin);
	    strcat (msgbuf, " bytes each to generate one output");
	    strcat (msgbuf, " line -- there is insufficient");
	    strcat (msgbuf, " memory to hold them."); 
	    c_errmsg (msgbuf, "regrid-intoobig", &nonfatal);
	    strcpy (msgbuf, "Divide output window into vertical"); 
	    strcat (msgbuf, " strips, REGRID each strip"); 
	    strcat (msgbuf, " separately, and seam them together");
	    strcat (msgbuf, " using CONCAT-AUTO.");
	    c_errmsg (msgbuf, "regrid-intoobig", &nonfatal);
	    ABORT;
	    }
	}
    free (pad);

/*  If entire input fits in memory, read it  */

    for (i = 0; i < nbndin; i++)
 	bndndx[i] = idsize * maxnli * nsi * i;
    badread = FALSE;
    if (maxnli == nli)
	{
	printf ("REGRID   Reading input image\n");
	for (i = 0; (i < nbndin) && !badread; i++)
	    {
	    eband = bands[i];
	    ebuf = inbuf + bndndx[i];
	    if (c_eread (&infd, &eband, &one, ebuf, &nli)
	      != maxnli * nsi)
	        badread = TRUE;
	    }
	flbuf = 0;
	llbuf = nli - 1;
	}
    else
	{
	flbuf = nli ;
	llbuf = 0;
	}
	  
/*  Get input coords of upper corners of each pixel on first output
    line  */

    rgd_incorn (olg[0], ncol, olg, osg, vertex, incorn);

/*  Initialize output min/max tally, if requested  */

    if (ifminmax)
	c_inmnmx (&odcode, &nbndout, *minval, *maxval);

/*  Start processing progress messages, if requested  */

    if (pmsgtim > 0)
	{
	if (nbndout == 1)
	    mtype = GEN2;
	else
	    mtype = ALLCAT;
	c_inittm ("REGRID", "generated", "output line", "bands ", " ",
	  &nlo, &nbndout, &pmsgtim, &mtype);
	}

/*  Get buffer offsets for first output line  */

    if (c_estep (&gdesc) != E_SUCC)
	ABORT;

/*  Generate one output line at a time  */

    for (oline = 0; oline < nlo; oline++)
	{
	if (maxnli != nli)
	    {

/*  Determine what additional input lines are needed for output line  */

	    inlim[0][0] = startil[oline];
	    inlim[0][1] = lastil[oline];
	    for (i = 0; i < 2; i++)
		{
		if (inlim[0][i] < 0) inlim[0][i] = 0;
		else if (inlim[0][i] >= nli) inlim[0][i] = nli - 1;
		}
	    wrap = 0;
	    if (inlim[0][1] % maxnli < inlim[0][0] % maxnli)
		{
		inlim[1][0] = maxnli * (inlim[0][1] / maxnli);
		inlim[1][1] = inlim[0][1];
		inlim[0][1] = inlim[1][0] - 1;
		wrap++;
		}
	    for (i = 0, nblkrd = 0; i <= wrap; i++)
		{
		if ((inlim[i][1] < flbuf) || (inlim[i][0] > llbuf))
		    {
		    blklim[nblkrd][0] = inlim[i][0];
		    blklim[nblkrd++][1] = inlim[i][1];
		    }
		else
		    {
		    if (inlim[i][0] < flbuf)
			{
			blklim[nblkrd][0] = inlim[i][0];
			blklim[nblkrd++][1] = flbuf - 1;
			}
		    if (inlim[i][1] > llbuf)
			{
			blklim[nblkrd][0] = llbuf + 1;
			blklim[nblkrd++][1] = inlim[i][1];
			}
		    }
		}

/*  Update values of first, last input lines in buffer  */

	    if (nblkrd > 0)
		{
		if (flbuf > inlim[0][0])
		    flbuf = inlim[0][0];
		else if (flbuf <= inlim[wrap][1] - maxnli)
		    flbuf = inlim[wrap][1] - maxnli + 1;
		if (llbuf < inlim[wrap][1])
		    llbuf = inlim[wrap][1];
		else if (llbuf >= inlim[0][0] + maxnli)
		    llbuf = inlim[0][0] + maxnli - 1;
		}

/*  If more input is needed, read it in  */

	    for (j = 0; (j < nblkrd) && !badread; j++)
	        {
		eline = blklim[j][0] + 1;
		nlblk = blklim[j][1] - blklim[j][0] + 1;
		for (i = 0; (i < nbndin) && !badread; i++)
		    {
		    eband = bands[i];
		    ebuf = inbuf + bndndx[i]
		      + (blklim[j][0] % maxnli) * nsi * idsize;
		    if (c_eread (&infd, &eband, &eline, ebuf, &nlblk)
		      != nlblk * nsi)
		        badread = TRUE;
		    }
		}
	    if (badread)
		{
		sprintf (msgbuf, "%s%s%ld%s%ld%s%ld", "Error reading",
		  " input image, ", inlim[j-1][0], " lines of band ",
		  eband, " starting at line ", eline);
		c_errmsg (msgbuf, "regrid-inread", &nonfatal);
		if (oline > 0)
		    {
		    sprintf (msgbuf, "Error occurred after %ld", oline);
		    strcat (msgbuf, " output lines were written");
		    c_errmsg (msgbuf, "informational", &nonfatal);
		    }
		ABORT;
		}
	    }

/*  Get input coords of lower corners of each output pixel on line  */	

	ulcptr = incorn + 2 * (nso + 1) * (oline % 2);
	llcptr = incorn + 2 * (nso + 1) * ((oline + 1) % 2);
	rgd_incorn ( 0.5 + (float) oline, ncol, olg, osg, vertex,
	  llcptr);

/*  Tally input contributions to build one output pixel at a time  */

	for (osamp = 0, obufbase = outbuf; osamp < nso;
	  osamp++, ulcptr += 2, llcptr += 2, obufbase += odsize)
	    {
	    for (i = 0; i < 2; i++)
		for (j = 0; j < 2; j++)
		    {
		    ovtex[i][j] = *(ulcptr + 2 * i + j);
		    ovtex[3-i][j] = *(llcptr + 2 * i + j);
		    }
	    rgd_tally (aggrcode, bndndx, fill, idcode, idsize, inbuf,
	      maxnli, nbndin, nbndout, nclass, nli, nsi, odcode, odsize,
	      offsets, ovtex, scalefact, wtr, accum, obufbase);
	    }

/*  Update output pixel minima/maxima if requested  */

	if (ifminmax)
	    for (i = 0; i < nbndout; i++)
		c_pxmnmx ((void *) (outbuf + odsize * offsets[i]), &nso,
		  *minval + i, *maxval + i, &odcode);

/*  Write output line  */

	if (c_estep (&gdesc) != E_SUCC)
	    ABORT;

/*  Update count for processing progress messages  */

	if (pmsgtim > 0)
	    {
	    curent = oline + 1;
	    if (oline == 0)
		c_startm ();
	    else if (curent == nlo)
		c_stoptm ();
	    }
	}


/*  Close image files  */

    free (inbuf);
    c_eclose (&infd);
    c_egclse (&gdesc);

    return;
    }			/* End of Function rgd_aggreg */
