/*  rgd_bldddr.c -- Build DDR for output image

NAME:
    rgd_bldddr ()

PURPOSE:
    Called by program regrid to enter the mapping data into the DDR for
    the output image and, if $MINMAX is set, insert the minimum and
    maximum values for each band.

ARGUMENTS:
    struct GEOGRID
	    *grid	(in)	Mapping grid.
    char    ifminmax	(in)	Set TRUE if $MINMAX = "yes".
    double  maxval[]	(in)	Array of maximum pixel values for each
				output band.
    double  minval[]	(in)	Array of minimum output values.
    long    nbndout	(in)	Number of bands in output image.
    char    outfile[]	(in)	Host name of output image file.
    long    owindow[4]	(in)	Output image window:  sl, ss, nl, ns.

RETURN VALUE:
    None.

FILE REFERENCES:
    <outfile>.DDR	Data descriptor record file, updated to reflect
			new geometry and min/max for each band.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:
    These are described in REGRID.PDF.

ALGORITHM DESCRIPTION:
    Retrieve skeleton DDR for output image <c_getddr>
    For each output band
	Save band-dependent DDR <c_getbdr>
    Copy geographic projection data from mapping grid file to DDR
    Compute projection coordinates of each corner of output image
    If simple change of scale
	Compute master image coordinates of upper left corner
    Write updated band-independent part of DDR <c_putddr>
    For each output band
	Write band-dependent DDR data <c_putddr>
    If min/max requested, insert values into ddr <c_svmnmx>
    									
***************   End of Prolog for rgd_bldddr   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<math.h>	/* Mathematical function definitions	*/
#include	<string.h>	/* String processing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"geompak.h"	/* Geom. transform package definitions	*/
#include	"proj.h"	/* Map projection codes */
#include	"regrid.h"	/* Definitions unique to REGRID		*/
									/*
***************   Start of Function rgd_bldddr   ***************	*/

FUNCTION void rgd_bldddr (
    struct GEOGRID
		*grid,		/* Mapping grid */
    char	ifminmax,	/* Set TRUE if $MINMAX = "yes" */
    double	*maxval,	/* Array of maximum pixel values for
				   each output band */
    double	*minval,	/* Array of minimum output values */
    long	nbndout,	/* Number of bands in output image */
    char	outfile[CMLEN],	/* Host name of output image file */
    long	owindow[4])	/* Output image window */
    {
	                                              			/*
Local Variable Definitions:						*/

    struct BDDR outbdr[MAXBND];	/* Band specific DDR record */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    float	grad;		/* Output coordinate gradient on edge of
				   of mapping grid */
    long	i;		/* Loop counter/index */
    float	master_x;	/* Master-image x-coord of output ulc */
    float	master_y;	/* Master-image y-coord of output ulc */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    struct DDR	outddr;		/* Buffer for output image DDR */

/*  Retrieve skeleton DDR for output image  */

    if (c_getddr (outfile, &outddr) != E_SUCC)
	{
	c_errmsg ("Unable to update output DDR", "regrid-noupddr",
	  &nonfatal);
	ABORT;
	}

/*  Save band-dependent DDR for each output band  */

    for (i = 1; i <= nbndout; i++)
	if (c_getbdr (outfile, &outbdr[i-1], &i) != E_SUCC)
	    {
	    c_errmsg ("Unable to update output DDR", "regrid-noupddr",
	      &nonfatal);
	    ABORT;
	    }

/*  Copy geographic projection data from mapping grid file to DDR  */

    for (i = 0; i < 8; i++)
	outddr.valid[i] = INVAL;      
    if ((outddr.valid[DDPCV] = grid->proj_valid) == VALID)
	{
	outddr.proj_code = grid->code;
	if ((grid->code == UTM) || (grid->code == SPCS))
	    {
	    outddr.zone_code = grid->zone;
	    outddr.valid[DDZCV] = VALID;
	    }
	outddr.datum_code = grid->datum;
	outddr.valid[DDDCV] = VALID;
	strcpy (outddr.proj_units, grid->units);
	outddr.valid[DDPUV] = VALID;
	for (i = 0; i < 15; i++)
	    outddr.proj_coef[i] = grid->projprms[i];
	outddr.valid[DDPPV] = VALID;
	outddr.pdist_y = grid->pdist[0];
	outddr.pdist_x = grid->pdist[1];
	outddr.valid[DDPDV] = VALID;

/*  Compute projection coordinates of each corner of output image  */

	grad = (grid->corners[4] - grid->corners[0])
	  / (owindow[2] - 1);
	outddr.upleft[0] = grid->corners[0] + (owindow[0] - 1) * grad;
	outddr.upright[0] = outddr.upleft[0] + (owindow[2] - 1) * grad;
	grad = (grid->corners[6] - grid->corners[2])
	  / (owindow[2] - 1);
	outddr.loleft[0] = grid->corners[2] + (owindow[0] - 1) * grad;
	outddr.loright[0] = outddr.loleft[0] + (owindow[2] - 1) * grad;
	grad = (grid->corners[3] - grid->corners[1])
	  / (owindow[3] - 1);
	outddr.upleft[1] = grid->corners[1] + (owindow[1] - 1) * grad;
	outddr.loleft[1] = outddr.upleft[1] + (owindow[3] - 1) * grad;
	grad = (grid->corners[7] - grid->corners[5])
	  / (owindow[3] - 1);
	outddr.upright[1] = grid->corners[5] + (owindow[1] - 1) * grad;
	outddr.loright[1] = outddr.upright[1] + (owindow[3] - 1) * grad;
	outddr.valid[DDCCV] = VALID;

/*  If simple change of scale, enter master image coordinates  */

	outddr.valid[DDINCV] = VALID;
	outddr.master_line = outddr.master_sample = 1;
	if ((grid->nrows == 2) && (grid->ncols == 2)
	    && (fabs (grid->in_lines[1] - grid->in_lines[0]) < .001)
	    && (fabs (grid->in_lines[3] - grid->in_lines[2]) < .001)
	    && (fabs (grid->in_samps[2] - grid->in_samps[0]) < .001)
	    && (fabs (grid->in_samps[3] - grid->in_samps[1]) < .001))
	    {
	    outddr.line_inc = (grid->in_lines[2] - grid->in_lines[0])
	      / (grid->lines - 1);
	    outddr.sample_inc = (grid->in_samps[1] - grid->in_samps[0])
	      / (grid->samples - 1);
	    master_y = grid->in_lines[0]
	      + outddr.pdist_y * (owindow[0] - 1) + .001;
	    master_x = grid->in_samps[0]
	      + outddr.pdist_x * (owindow[1] - 1) + .001;
	    if ((master_y - floor (master_y) < .002)
		&& (master_x - floor (master_x) < .002))
		{
		outddr.master_line = master_y;
		outddr.master_sample = master_x;
		}
	    else
		outddr.master_line = outddr.master_sample = 0;
	    }
	}

/*  Write updated band-independent part of DDR  */

    if (c_putddr (outfile, &outddr) != E_SUCC)
	{
	c_errmsg ("Unable to write output DDR", "regrid-noupddr",
	  &nonfatal);
	return;
	}

/*  Write previously saved band-dependent DDR records  */

    for (i = 0; i < nbndout; i++)
	{
	if (c_putbdr (outfile, &outbdr[i]) != E_SUCC)
	    {
	    c_errmsg ("Unable to write output BDDR", "regrid_noupbddr",
	      &nonfatal);
	    return;
	    }
	}

/*  If min/max requested, insert values into ddr  */

    if (ifminmax)
	if (c_svmnmx (outfile, &nbndout, minval, maxval, "YES")
	  != E_SUCC)
	    c_errmsg ("Unable to enter min/max values",
	      "regrid-noupbddr", &nonfatal);

    return;
    }			/* End of Function rgd_bldddr	*/
