/*  RGD_CKDDR.C -- Check input DDR, getting classes and output bands

NAME:
    rgd_ckddr ()

PURPOSE:
    Called by regrid to get the DDR for the input image, check that the
    requested aggregation mode is compatible with the image data type,
    and then determine the maximum value of any class and the number of
    output bands.

ARGUMENTS:
    long    aggrcode	(in)	Code for aggregation mode.
    char    *infile	(in)	Name of input image file.
    long    bands[]	(in)	List of input bands to use.
    long    nbands	(in)	Number of input bands to use.
    struct DDR
	    *inddr	(out)	DDR for input image.
    long    *nclass	(out)	Number of classes.
    long    *nbndout	(out)	Number of bands in output image.

RETURN VALUE:
    None.

FILE REFERENCES:
    infile.DDR		(in)	DDR for input image; refereced to
				determine image data type and maximum
				pixel value.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in regrid.pdf.

ALGORITHM DESCRIPTION:
    Get input image DDR <c_getddr>
    Get input data type
    If aggregation function is MODE
	If single input channel
	    If input data type is real
		Advise user to map data to integer values
		Abort processing
	    Else
		Set flag for checking BDR
	Else
	    Set number of classes to number of input bands + 1
	Set number of output bands to one
    Else if aggregation function is MEAN
	Set number of classes to number of input bands
	Set number of output bands to number of input bands
    Else (aggregation function is FRACT)
	If number of bands > 1
	    Inform user that FRACT requires a single input band
	    Abort
	If input data type is real
	    Advise user to map data to integer values
	    Abort processing
	Else
    	    Set flag for checking BDR
    If BDR flag is set
	Get band dependent DDR record for input band <c_getbdr>
	If maximum value for input band not specified
	    If function MODE with datatype BYTE
		Set number of classes to 256
	    Else
		Advise user to run program MINMAX
		Abort
	Else
	    Set number of classes to maximum input value + 1
	If minimum input value < 1
	    Inform user that these modes require non-negative input
	    Abort
	If aggregation function is FRACT
	    Set number of output bands to maximum value
	    If maximum value exceeds maximum allowed bands in one image 
		Advise user to compress data range
		Abort
	    If minimum input value is less than 1
		Advise user that fractional coverage for values less
		  than 1 will not be recorded in output image
    									
***************   End of Prolog for rgd_ckddr   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String processing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"regrid.h"	/* Regrid-specific definitions		*/
									/*
***************   Start of Function rgd_ckddr   ***************	*/

FUNCTION void rgd_ckddr (
    long	aggrcode,	/* Code for aggregation mode */
    char	 *infile,	/* Name of input image file */
    long	 bands[],	/* List of input bands to use */
    long	 nbands,	/* Number of input bands to use */
    struct DDR	*inddr,		/* DDR for input image */
    long	*nclass,	/* Number of classes */
    long	*nbndout)	/* Number of bands in output image */
    {
	                                              			/*
Local Variable Definitions:						*/

    struct BDDR bddr;		/* Band specific DDR record */
    long	ckbdr;		/* Flag set if BDR needs to be read */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    char	msgbuf[128];	/* Buffer for building error message */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */

/*  Get input image DDR  */

    if (c_getddr (infile, inddr) != E_SUCC)
	{
	c_errmsg ("DDR for input image not found", "regrid-noddr",
	  &nonfatal);
	ABORT;
	}
    ckbdr = FALSE;

/*  Check MODE aggregation function for valid data type  */

    if (aggrcode == RGD_MODE)
	{
	if (nbands == 1)
	    {
	    if (inddr->dtype > 3)
		{
		sprintf (msgbuf, "Input image data type must %s",
		  "be integer (BYTE, I*2, or I*4)");
		c_errmsg (msgbuf, "regrid-baddtype", &nonfatal);
		ABORT;
		}
	    else
		ckbdr = TRUE;
	    }
	else
	    *nclass = nbands + 1;
	*nbndout = 1;
	}

/*  For aggregation mode MEAN, no special checks are needed  */

    else if (aggrcode == RGD_MEAN)
	{
	*nclass = nbands;
	*nbndout = nbands;
	}

/*  For FRACT aggregation type, check that only single band specified */

    if (aggrcode == RGD_FRACT)
	{
	if (nbands != 1)
	    {
	    strcpy (msgbuf,
	      "Aggregation mode FRACT accepts only a single input");
	    sprintf (msgbuf + strlen (msgbuf),
		  " band, with maximum value %d", MAXBND);
	    c_errmsg (msgbuf, "regrid-fractband", &nonfatal);
	    ABORT;
	    }

/*  Also check that input is integer data type  */

	if (inddr->dtype > 3)
	    {
	    sprintf (msgbuf, "Input image data type must %s",
	      "be integer (BYTE, I*2, or I*4)");
	    sprintf (msgbuf + strlen (msgbuf),
	      " with maximum value %d", MAXBND);
	    c_errmsg (msgbuf, "regrid-baddtype", &nonfatal);
	    ABORT;
	    }
	else
	    ckbdr = TRUE;
	}

/*  If necessary, get maximum input pixel value from BDR  */

    if (ckbdr)
	{
	if (c_getbdr (infile, &bddr, bands) != E_SUCC)
	    {
	    c_errmsg ("Error reading band-specific DDR record",
	      "regrid-nobdr", &nonfatal);
	    ABORT;
	    }

/*  Check for acceptable maximum value  */

	if (bddr.valid != VALID)
	    {
	    if ((aggrcode == RGD_MODE) && (inddr->dtype == EBYTE))
		*nclass = 256;
	    else
		{
		strcpy (msgbuf, "Maximum pixel value not in DDR --");
		strcat (msgbuf, " run program MINMAX on input image");
		c_errmsg (msgbuf, "regrid-nomaxval", &nonfatal);
		ABORT;
		}
	    }
	else
	    if (bddr.minval < -.0001)
		{
		strcpy (msgbuf, "Input has negative values -- REGRID");
		strcat (msgbuf, " cannot process them.  Rescale data");
		strcat (msgbuf, " (e.g., using program MAP) and rerun");
		strcat (msgbuf, " REGRID");
		c_errmsg (msgbuf, "regrid-neginval", &nonfatal);
		ABORT;
		}
	    *nclass = (long) bddr.maxval + 1;

/*  Check that maximum does not exceed allowed number of bands  */

	if (aggrcode == RGD_FRACT)
	    {
	    *nbndout = *nclass - 1;
	    if (*nbndout > MAXBND)
		{
		sprintf (msgbuf, "%s%ld%s%d%s%s",
		   "Maximum input pixel value ", *nbndout,
		  " exceeds maximum number of output bands ",
		  MAXBND, " -- compress data range (e.g., using ",
		  "program MAP) and rerun REGRID");
		c_errmsg (msgbuf, "regrid-pxltoobig", &nonfatal);
		ABORT;
		}

/*  If input has values less than 1, warn user they will be ingnored  */

	    if ( (long) bddr.minval == 0)
		{
		strcpy (msgbuf, "Input pixels with value 0 will");
		strcat (msgbuf, " be ignored when computing");
		strcat (msgbuf, " fractional coverage");
		c_errmsg (msgbuf, "regrid-zerovals", &nonfatal);
		}
	    }
	}

/*  Return to caller  */

    return;
    }			/* End of Function rgd_ckddr  */
