/*  rgd_getarg -- Ingest program arguments

NAME:
    rgd_getarg ()

PURPOSE:
    This function is called by program REGRID to get program arguments
    from the TAE interface.

ARGUMENTS:
    The arguments for REGRID are described in the regrid.c header.

    Arguments for function rgd_getarg are:

    struct PARBLK
	    *vblock	(in)	TAE parameter block.
    long    *aggrcode	(out)	Aggregation function to be performed.
    char    *infile	(out)	Name of input image file.
    long    bands[]	(out)	Input image bands to use.
    long    *nbands	(out)	Number of input bands to use.
    long    iwindow[4]	(out)	Input window specifications, if given.
    char    *gridfile	(out)	Name of input mapping grid file.
    char    *outfile	(out)	Name of output image file.
    long    *windopt	(out)	Code for how output window is defined.
    long    *owindow	(out)	Output window specification, if given.
    float   *odcode	(out)	LAS data type code of output image.
    float   *fill	(out)	Fill value for pixels outside input.
    struct WTR
	    *wtr	(out)	Class weight, threshold, and
				reallocation data.
    char    *ifminmax	(out)	Set TRUE if $MINMAX variable = "yes".
    long    *pmsgtim	(out)	TAE $MSGTIM variable.

RETURN VALUE:
    None.

PROGRAM HISTORY:                                  
    1993-11	Initial version, R. A. White, PSU/ORSER.
    1994-07	Added support for class reallocation, R. A. White,
	   	PSU/ESSC.
    1994-12	Added support for Unix pathname syntax, raw.
    1999-02	Added subcommands to specify aggregation mode, raw.
    2001-08	Fixed test for whether grd_getwtr() call needed, raw.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in RGD_GETARG.PDF.

EXTERNAL FUNCTIONS CALLED:

(Program Specific Functions)
    rgd_getwtr		Get class weight, threshold, and reallocation
			arguments and data.

ALGORITHM DESCRIPTION:
    Get subcommand name from TAE interface
    Set aggregation function code based on subcommand
    Get input filename, bands, and window
    Get input grid filename
    Get output filename, checking for already existing
    Get windowing option and convert to numeric code
    If output window is specified
	Get output window values
    Get output data type and convert to numeric code
    Get fill value for output pixels entirely outside input image
    If class weighting, thresholding, and/or reallocation is specified
	Get weights, thresholds, and reallocations <rgd_getwtr>
    Get system variables for message time and min/max option
    									
***************   End of Prolog for rgd_getarg   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String processing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"alias.h"	/* LAS environment variable function	*/
#include	"labserv.h"	/* LAS label services functions		*/
#include	"pixman.h"	/* LAS pixel maniputaliton functions	*/
#include	"util.h"	/* LAS utility functions		*/
#include	"regrid.h"	/* Regrid-specific definitions		*/
 									/*
External Function Prototypes Not Included in Header Files:		*/

    struct VARIABLE *p_find  (struct PARBLK *vblock, char *vname);
    struct VARIABLE *p_fvar  (struct PARBLK *vblock, char *vname);
    void	    rgd_getwtr (struct PARBLK *vblock, struct WTR *wtr);
									/*
***************   Start of Function rgd_getarg   ***************	*/

FUNCTION void rgd_getarg (
    struct PARBLK
	    *vblock,		/* TAE parameter block */
    long    *aggrcode,		/* Aggregation function to be performed */
    char    *infile,		/* Name of input image file */
    long    bands[],		/* Input image bands to use */
    long    *nbands,		/* Number of input bands to use */
    long    iwindow[4],		/* Input window specifications, if given */
    char    *gridfile,		/* Name of input mapping grid file */
    char    *outfile,		/* Name of output image file */
    long    *windopt,		/* Code for how output window is defined */
    long    *owindow,		/* Output window specification, if given */
    long    *odcode,		/* LAS data type code of output image */
    float   *fill,		/* Fill value for pixels outside input */
    struct WTR
	    *wtr,		/* Class weight, threshold, and
				   reallocation data */
    char    *ifminmax,		/* Set TRUE if $MINMAX variable = "yes" */
    long    *pmsgtim)		/* TAE $MSGTIM variable */
    {

/* Local Variable Definitions:						*/

    struct VARIABLE
		*argptr;	/* Pointer to TAE argument list	*/
    long	dummy[2];	/* Dummy argument for calls */
    char	argbuf[12];	/* Value of string-type argument */
    long	dtype;		/* Data type for pxcopy () */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    long	i;		/* Loop index */
    char	msgbuf[ERRLEN];	/* Buffer for error messages */
    long	namlng;		/* Length of name string */
    long	ncopy;		/* Number of bytes to copy */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    long	one = 1;	/* Constant one */
    TAEFLOAT	real;		/* Temporary storage for real argument */
    char	tname[CMLEN];	/* TAE name of file (not checked) */
    long	zero = 0;	/* Constant zero */

/*  Get aggregation function name from TAE interface and set code  */

    strcpy(argbuf, SVAL(*p_find(vblock, "_SUBCMD"), 0));
    if (argbuf[1] == 'O')
	*aggrcode = RGD_MODE;
    else if (argbuf[1] == 'E')
	*aggrcode = RGD_MEAN;
    else
	*aggrcode = RGD_FRACT;

/*  Get input filename, bands, and window  */

    if (c_getin (vblock, "IN", &zero, (void *) tname, (void *) infile,
      dummy, (void *) iwindow, (void *) bands, nbands, dummy, &one,
      dummy, &dummy) != E_SUCC)
	ABORT;

/*  Get input grid filename and check for existence  */

    strcpy (gridfile, SVAL(*p_find (vblock, "GRID"), 0));
    c_getalias (gridfile, gridfile);
    namlng = strlen(gridfile);
    strcat (gridfile, ".grid");
    gridfile[namlng] = HOST_EXT;
    if (c_lsstat (gridfile, &zero) != E_SUCC)
	{
	gridfile[namlng] = '\0';
	if (c_lsstat (gridfile, &zero) != E_SUCC)
	    {
	    sprintf (msgbuf, "Cannot find grid file \"%s\"", gridfile);
	    c_errmsg (msgbuf, "regrid-nogridfil", &nonfatal);
	    ABORT;
	    }
	}

/*  Get output filename, checking for already existing  */

    if (c_getout (vblock, "OUT", (void *) tname, (void *) outfile) !=
      E_SUCC)
	ABORT;

/*  Get windowing option and convert to numeric code  */

    strcpy(argbuf, SVAL(*p_find(vblock, "WINDOPT"), 0));
    if (argbuf[0] == 'I')
	*windopt = WOP_IN;
    else
	{
	*windopt = WOP_OUT;

/*  Get output window values  */

	argptr = p_find (vblock, "OWINDOW");
	owindow[2] = -1;
	for (i = 0; i < argptr->v_count; i++)
	    owindow[i] = IVAL (*argptr, i);
	}

/*  Get output data type and convert to numeric code  */

    strcpy(argbuf, SVAL(*p_find(vblock, "ODTYPE"), 0));
    if (strcmp (argbuf, "BYTE") == 0)
	*odcode = EBYTE;
    else if (strcmp (argbuf, "I*2") == 0)
	*odcode = EWORD;
    else if (strcmp (argbuf, "I*4") == 0)
	*odcode = ELONG;
    else if (strcmp (argbuf, "R*4") == 0)
	*odcode = EREAL;
    else
	*odcode = 0;		/* Default to same as input */

/*  Get fill value for output pixels entirely outside input  */

    dtype = EBYTE;
    ncopy = sizeof (TAEFLOAT);
    c_pxcopy ( (*p_find (vblock, "FILL")).v_cvp, (char *) &real, &dtype,
      &ncopy);
    *fill = real;
    if (*aggrcode == RGD_FRACT)
	*fill = 0.0;

/*  Get NODATA value for subcommand MEAN  */

    if ( (*aggrcode == RGD_MEAN)
      && ( (*p_find (vblock, "NODATA")).v_count > 0) )
	{
	wtr->ndflag = 1;
	wtr->nodata = RVAL(*p_find (vblock, "NODATA"), 0);
	}
    else
	wtr->ndflag = 0;

/*  Get weights, thresholds, and/or reallocations, if any  */

    wtr->wcount = wtr->tcount = wtr->rcount = 0;
    if ( (p_fvar (vblock, "WCLASS") != NULL)
      || (p_fvar (vblock, "TCLASS") != NULL)
      || (p_fvar (vblock, "REALLOC") != NULL) )
	rgd_getwtr (vblock, wtr);

/*  Get system variables for message time and min/max option  */

    *pmsgtim = IVAL(*p_find(vblock, "$MSGTIM"), 0);
    strcpy(argbuf, SVAL(*p_find(vblock, "$MINMAX"), 0));
    if (argbuf[0] == 'Y')
        *ifminmax = TRUE;
    else
        *ifminmax = FALSE;

    return;
    }			/* End of Function rgd_getarg	*/
