/*  RGD_GETWTR.C -- Get class weights, thresholds, and reallocations

NAME:
    reg_getwtr ()

PURPOSE:
    Called by rgd_getarg as part of program REGRID to get arguments
    and/or file specifying class weights, thresholds, and reallocations.

ARGUMENTS:
    struct PARBLK *vblock   (in)    TAE parameter block.
    struct WTR	  *wtr	    (out)   Class weight, threshold, and
				    reallocation data.

RETURN VALUE:  None.

FILE REFERENCES:
    <REALLOC>	READ access to file specified by argument REALLOC to
		obtain table of class reallocations, weights, and/or
		thresholds.

PROGRAM HISTORY:                                  
    1994-07	Initial version, R. A. White, PSU/ESSC.
    1994-12	Added support for Unix pathname syntax, raw.
    2001-08	Added support for subcommand-dependent args, raw.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.8 and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in REGRID.PDF.

ALGORITHM DESCRIPTION:
    Allocate initial space for weight and threshold arrays
    If reallocation file is specified
	Open reallocation file
	Allocate initial space for allocation data
	For each record in allocation file
	    If record is reallocation record
		Extract oldclass, newclass, and weight values
		Update counts of old and new classes, allocating more
		  table space if needed
	    Else if record is weight record
		Extract class and weight values
		Update weight count, extending table space if needed
	    Else if record is threshold record
		Extract class and threshold values
		Update threshold count, extending table space if needed
	    Else
		Ignore record
	Close reallocation file
    If weights entered directly as arguments
	For each class in argument list
	    If class is already in weight class table
		Replace previous weight value for class with new value
	    Else
		Add new class and weight value to table
    If thresholds entered directly as arguments
	For each class in argument list
	    If class is already in threshold class table
		Replace previous threshold value with new value
	    Else
		Add new class and threshold value to table
    Return
    									
***************   End of Prolog for rgd_getwtr ()   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String function definitions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"alias.h"	/* LAS env. variable function		*/
#include	"pixman.h"	/* LAS pixel manipulation functions	*/
#include	"util.h"	/* LAS utility functions		*/
#include	"regrid.h"	/* Program-specific definitions		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    struct VARIABLE *p_find  (struct PARBLK *vblock, char *vname);
    struct VARIABLE *p_fvar  (struct PARBLK *vblock, char *vname);
									/*
***************   Start of Function rgd_getwtr ()   ***************	*/

FUNCTION void rgd_getwtr (
    struct PARBLK
		*vblock,	/* TAE parameter block */
    struct WTR	*wtr)		/* Class weight, threshold, and
				   reallocation data */
    {
	                                              			/*
Local Variable Definitions:						*/

    struct VARIABLE
		*argptr;	/* Pointer to TAE argument list */
    struct VARIABLE
		*classptr;	/* Pointer to class numbers in TAE
				   argument list */
    long	dtype = EBYTE;	/* Data type for c_pxcopy */
    long	errcnt;		/* Number of bad reallocation records */
    long	fatal = -1;	/* Fatal error flag for c_errmsg () */
    long	i;		/* Loop counter/index */
    char	inbuf[ERRLEN];	/* Buffer for reallocation file input */
    long	index;		/* Index into input buffer */
    long	j;		/* Loop counter/index */
    long	nclass;		/* Number of classes in TAE arg. list */
    long	newclalloc;	/* Space allocated for new classes */
    long	newclcnt;	/* Number of new classes */
    long	nonfatal = 2;	/* Warning status flag for c_errmsg () */
    long	oldclass;	/* Old class from R record */
    long	rcalloc;	/* Space allocated for reallocation
				   classes */
    TAEFLOAT	real;		/* Temporary storage for real argument */
    char	reallocfile[CMLEN]; /* Name of reallocation file */
    FILE	*rfptr;		/* File pointer for reallocation file */
    long	rsize		/* Length of TAE real argument */
		  = sizeof (TAEFLOAT);
    long	thrshalloc;	/* Space allocated for thresholds */
    long	wghtalloc;	/* Space allocated for weights */

/*  Allocate weight and threshold arrays  */

    rcalloc = thrshalloc = wghtalloc = 10;
    wtr->wclass = (long *) malloc (wghtalloc * sizeof (long));
    wtr->weight = (float *) malloc (wghtalloc * sizeof (float));
    wtr->tclass = (long *) malloc (thrshalloc * sizeof (long));
    wtr->thresh = (float *) malloc (thrshalloc * sizeof (float));

/*  Determine whether reallocation file is to be used  */

    argptr = p_fvar (vblock, "REALLOC");
    if ((argptr != NULL) && (argptr->v_count > 0))
	{

/*  Open reallocation file  */

	strcpy (reallocfile, SVAL (*argptr, 0));
	c_getalias (reallocfile, reallocfile);
	if ((rfptr = fopen (reallocfile, "r")) == NULL)
	    {
	    perror ("[regrid-realopen]");
	    c_errmsg ("Error occurred opening reallocation file",
	      "regrid_realopen", &nonfatal);
	    ABORT;
	    }

/*  Process records according to type  */

	newclalloc = 20;
	wtr->rclass = (long *) malloc (rcalloc * sizeof (long));
	wtr->nclcnt = (long *) malloc (rcalloc * sizeof (long));
	wtr->newcl = (long *) malloc (newclalloc * sizeof (long));
	wtr->fract = (float *) malloc (newclalloc * sizeof (float));
	newclcnt = 0;
	errcnt = 0;
	oldclass = - 987654321;
	while (fgets (inbuf, ERRLEN, rfptr) != NULL)
	    {
	    if ((index = strcspn (inbuf, "rRtTwW"))
	      < (long) strspn (inbuf, "rRtTwW \t"))
		{
		if ((inbuf[index] == 'r') || (inbuf[index] == 'R'))

/*  Reallocation Record  */

		    {
		    if (sscanf (inbuf + ++index, "%ld %ld %f",
		      wtr->rclass + wtr->rcount, wtr->newcl + newclcnt,
		      wtr->fract + newclcnt) != 3)
			{
			c_errmsg ("Bad record in reallocation file:",
			  "regrid-realrec", &nonfatal);
			c_errmsg (inbuf + index - 1, "regrid-realrec",
			  &nonfatal);
			c_errmsg (" -- record ignored",
			  "regrid-realrec", &nonfatal);
			if (errcnt++ > 10)
			    {
			    c_errmsg ("Too many bad records",
			      "regrid-manybad", &nonfatal);
			    ABORT;
			    }
			}

/*  Update counts, allocating more table space as needed  */

		    if (oldclass != (wtr->rclass)[wtr->rcount])
			{
			oldclass = (wtr->rclass)[wtr->rcount];
			(wtr->nclcnt)[wtr->rcount] = 1;
			if ( (wtr->rcount)++ >= rcalloc)
			    {
			    rcalloc *= 2;
			    wtr->rclass = (long *) realloc (wtr->rclass,
			      rcalloc * sizeof (long));
			    wtr->nclcnt = (long *) realloc (wtr->nclcnt,
			      rcalloc * sizeof (long));
			    }
			}
		    else
			(wtr->nclcnt)[wtr->rcount - 1]++;
		    if (newclcnt++ >= newclalloc)
			{
			newclalloc *= 2;
			wtr->newcl = (long *) realloc (wtr->newcl,
			  newclalloc * sizeof (long));
			wtr->fract = (float *) realloc (wtr->fract,
			  newclalloc * sizeof (float));
			}
		    
		    }

/*  Weight record  */

		else if ((inbuf[index] == 'w') || (inbuf[index] == 'W'))
		    {
		    if (sscanf (inbuf + ++index, "%ld %f",
		      wtr->wclass + wtr->wcount,
		      wtr->weight + wtr->wcount) != 2)
			{
			c_errmsg ("Bad record in reallocation file:",
			  "regrid-realrec", &nonfatal);
			c_errmsg (inbuf + index - 1, "regrid-realrec",
			  &nonfatal);
			c_errmsg (" -- record ignored",
			  "regrid-realrec", &nonfatal);
			if (errcnt++ > 10)
			    {
			    c_errmsg ("Too many bad records",
			      "regrid-manybad", &nonfatal);
			    ABORT;
			    }
			}

/*  Update weight count, allocating more table space as needed  */

		    if ( (wtr->wcount)++ >= wghtalloc)
			{
			wghtalloc *= 2;
			wtr->wclass = (long *) realloc (wtr->wclass,
			  wghtalloc * sizeof (long));
			wtr->weight = (float *) realloc (wtr->weight,
			  wghtalloc * sizeof (float));
			}
		    }

/*  Threshold record  */

		else if ((inbuf[index] == 't') || (inbuf[index] == 'T'))
		    {
		    if (sscanf (inbuf + ++index, "%ld %f",
		      wtr->tclass + wtr->tcount,
		      wtr->thresh + wtr->tcount) != 2)
			{
			c_errmsg ("Bad record in reallocation file:",
			  "regrid-realrec", &nonfatal);
			c_errmsg (inbuf + index - 1, "regrid-realrec",
			  &nonfatal);
			c_errmsg (" -- record ignored",
			  "regrid-realrec", &nonfatal);
			if (errcnt++ > 10)
			    {
			    c_errmsg ("Too many bad records",
			      "regrid-manybad", &nonfatal);
			    ABORT;
			    }
			}

/*  Update threshold count, allocating more table space as needed  */

		    if ( (wtr->tcount)++ >= thrshalloc)
			{
			thrshalloc *= 2;
			wtr->tclass = (long *) realloc (wtr->tclass,
			  thrshalloc * sizeof (long));
			wtr->thresh = (float *) realloc (wtr->thresh,
			  thrshalloc * sizeof (float));
			}
		    }
		}
	    }

/*  Close file after last record processed  */

	fclose (rfptr);
	}
    
/*  Check whether weights were entered directly as arguments  */

    classptr = p_fvar (vblock, "WCLASS");
    if ((classptr != NULL) && ((nclass = classptr->v_count) > 0))
	{
	argptr = p_find (vblock, "WEIGHT");
	if (argptr->v_count != nclass)
	    {
	    c_errmsg (
	      "Number of WEIGHTs not same as number of classes",
	      "regrid-badwgtcnt", &nonfatal);
	    ABORT;
	    }
	while (wtr->wcount + nclass > wghtalloc)
	    {
	    wghtalloc *= 2;
	    wtr->wclass = (long *) realloc (wtr->rclass,
	      wghtalloc * sizeof (long));
	    wtr->weight = (float *) realloc (wtr->weight,
	      wghtalloc * sizeof (float));
	    }

/*  Read weights from TAE arguments, checking for duplicated class  */

	for (i = 0; i < nclass; i++)
	    {
	    wtr->wclass[wtr->wcount] = IVAL (*classptr, i);
	    for (j = 0; j < wtr->wcount; j++)
		if (wtr->wclass[wtr->wcount] == wtr->wclass[j])
		    break;
	    c_pxcopy ( (*argptr).v_cvp + i * rsize, (char *) &real,
	      &dtype, &rsize);
	    wtr->weight[j] = real;
	    if (j == wtr->wcount)
		wtr->wcount++;
	    }
	}

/*  Check whether thresholds were entered directly as arguments  */

    classptr = p_fvar (vblock, "TCLASS");
    if ((classptr != NULL) && ((nclass = classptr->v_count) > 0))
	{
	argptr = p_find (vblock, "THRESH");
	if (argptr->v_count != nclass)
	    {
	    c_errmsg (
	      "Number of THRESHolds not same as number of classes",
	      "regrid-badthrcnt", &nonfatal);
	    ABORT;
	    }
	while (wtr->tcount + nclass > thrshalloc)
	    {
	    thrshalloc *= 2;
	    wtr->tclass = (long *) realloc (wtr->tclass,
	      thrshalloc * sizeof (long));
	    wtr->thresh = (float *) realloc (wtr->thresh,
	      thrshalloc * sizeof (float));
	    }

/*  Read thresholds from arguments, checking for duplicated class  */

	for (i = 0; i < nclass; i++)
	    {
	    wtr->tclass[wtr->tcount] = IVAL (*classptr, i);
	    for (j = 0; j < wtr->tcount; j++)
		if (wtr->tclass[wtr->tcount] == wtr->tclass[j])
		    break;
	    c_pxcopy ( (*argptr).v_cvp + i * rsize, (char *) &real,
	      &dtype, &rsize);
	    wtr->thresh[j] = real;
	    if (j == wtr->tcount)
		wtr->tcount++;
	    }
	}

/*  Return to caller  */

    return;
    }			/* End of Function rgd_getwtr () */
