/*  rgd_incorn.c -- Get input coords for line of output pixels

NAME:
    rgd_incorn ()

PURPOSE:
    Called by rgd_aggreg as part of program REGRID to determine the
    input coordinates at the left and right edges of each output pixel
    along a separation line between two rows of output pixels.

ARGUMENTS:

    float   olcoord	(in)	Output line coordinate for which input
				coordinates are to be found.
    long     ncol	(in)	Number of columns in mapping grid.
    float    olg[]	(in)	Output line number on each grid row.
    float    osg[]	(in)	Output sample on each grid column.
    struct VTEX
	     vertex[]	(in)	Array of input coords and derivatives
				at grid vertices.
    float   incorn[]	(out)	Input (l,s) for each edge in turn.

    Note:  The mapping grid needs to have been trimmed to just cover the
    output window, extended outwards 0.5 pixel from the centers of the
    edge pixels.

RETURN VALUE:
    None.

PROGRAM HISTORY:                                  
    12-93	Initial version, R. A. White, PSU/ORSER.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Sun Sparc SunOS 4.1.3.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in REGRID.PDF.

ALGORITHM DESCRIPTION:
    Determine mapping grid row containing output line
    For each boundary between output pixels on this line
	If first pixel in new cell
	    Compute input x,y at left edge of cell
	    Compute x gradient of input x,y
	    Compute input x,y of first pixel
	Else
	    Add gradients to previous input x,y to get new values
    									
***************   End of Prolog for PROGRAM-NAME   ***************	*/
									/*
Header Files:			       					*/

#include 	"las.h"		/* Standard LAS include files */
#include	"regrid.h"	/* REGRID-unique definitiions */
									/*
***************   Start of Function rgd_incorn   ***************	*/

void FUNCTION rgd_incorn (
    float	olcoord,	/* Output line coordinate for which
				   input coordinates are to be found */
    long	ncol,		/* Number of columns in mapping grid */
    float	olg[],		/* Output line number on each grid row */
    float	osg[],		/* Output sample on each grid column */
    struct VTEX	vertex[],	/* Array of input coords and derivatives
				   at grid vertices */
    float	incorn[])	/* Input (l,s) for each edge in turn */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	col;   		/* Current mapping grid column */
    float	deltal;		/* Distance from grid row to current
				   output line */
    float	dlds;		/* dl(in) / ds(out) on output line */
    float	dsds;		/* ds(in) / ds(out) on output line */
    long	i;		/* Index into incoord array */
    float	leftl;		/* Input line number on output line at
				   left edge of mapping grid cell */
    float	lefts;		/* Input sample number on output line at
				   left edge of mapping grid cell */
    float	os;		/* Output sample coordinate */
    long	vtex;		/* Current grid vertex */
    long	row;		/* Current mapping grid row */


/*  Determine mapping grid row containing output line  */

    for (row = 0; olcoord > olg[row + 1]; row++);
    deltal = olcoord - olg[row];
    col = 0;
    vtex = row * ncol;
    leftl = vertex[vtex].il + vertex[vtex].dldl * deltal;
    dlds = vertex[vtex].dlds + vertex[vtex].d2l * deltal;
    lefts = vertex[vtex].is + vertex[vtex].dsdl * deltal;
    dsds = vertex[vtex].dsds + vertex[vtex].d2s * deltal;
    incorn[0] = leftl;
    incorn[1] = lefts;

/*  Find mapping grid column for each output pixel boundary  */

    for (os = osg[0] + 1.0, i = 1; os <= osg[ncol - 1]; os += 1.0, i++)
	{
	if (os > osg[col + 1])
	    {
	    col++;
	    vtex++;
	    leftl = vertex[vtex].il + vertex[vtex].dldl * deltal;
	    dlds = vertex[vtex].dlds + vertex[vtex].d2l * deltal;
	    lefts = vertex[vtex].is + vertex[vtex].dsdl * deltal;
	    dsds = vertex[vtex].dsds + vertex[vtex].d2s * deltal;
	    incorn[2*i] = leftl + dlds * (os - osg[col]);
	    incorn[2*i+1] = lefts + dsds * (os - osg[col]);
	    }
	else
	    {
	    incorn[2*i] = incorn[2*i-2] + dlds;
	    incorn[2*i+1] = incorn[2*i-1] + dsds;
	    }
	}

/*  Return to caller  */

    return;
    }			/* End of Function rgd_incorn	*/
