/*  rgd_inrect.c -- Determine input rectanagle for each output line

NAME:
    rgd_inrect ()

PURPOSE:
    Called by program REGRID to determine the rectangle of input data
    needed to generate each line of the output image.  This function
    creates a table of first and last input lines for each output line,
    and computes the maximum number of input lines needed by any output
    line.  If the region to be processed was specified in terms of the
    output window, rgd_inrect also computes the required input window.

ARGUMENTS:
    long    nlo		(in)	Number of lines in output image.
    long    nrow	(in)	Number of rows in mapping grid.
    long    ncol	(in)	Number of columns in mapping grid.
    float   olg[]	(in)	Output line number on each grid row.
    float   osg[]	(in)	Output sample on each grid column.
    struct VTEX
	    v[]		(in)	Array of input coords and derivatives
				at grid vertices.
    long    windopt	(in)	Window specification-option code.
    long    iwindow[4]	(mod)	Input window:  sl, ss, nl, ns.
    long    *maxnli	(out)	Maximum input lines for any output line.
    long    **startil	(out)	Starting input line for each output
				line.
    long    **lastil	(out)	Last input line for each output line.

    Note:  The mapping grid data (nrow, ncol, olg, osg, v) must have
    already been trimmed to a region 0.5 pixels larger on each edge than
    the output image window.

RETURN VALUE:
    None.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

ALGORITHM DESCRIPTION:
    Allocate arrays for intermediate data and final results
    Initialize minimum, maximum input line and pixel for entire image
    For each line separating rows of output pixels (line +/- 0.5)
	Determine mapping grid row containing this line
	If first output line in this grid row
	    Initialize input x,y for one line above current separator
	For each vertical line in grid
	    Increment input x,y for this vertical line
	    If first vertical line
		Initialize input min/max from x,y values
	    Else
		Update input min/max using x,y values <rgd_ckmnmx>
	Update input min/max for entire image from min/max for line
	Save minimum and maximum input lines for this separating line
    For each output line
        Combine input extrema along upper and lower separating lines to
	  compute first, last input lines for this output line
	Keep running maximum of number of input lines for one output
	  line
    If window was specified in terms of output image
	Determine needed input window
	If input extends outside available data
	    Print warning message
	Print adjusted input coordinates
    Make all input coordinates relative to window upper left corner
    									
***************   End of Prolog for rgd_inrect ()   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Math library definitions		*/
#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String prpcessing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"regrid.h"	/* Definitions unique to REGRID		 */
									/*
External Function Prototypes Not Included in Header Files:		*/
    void rgd_ckmnmx (float y, float x, float minmax[2][2]);
									/*
***************   Start of Function rgd_inrect   ***************	*/

void FUNCTION rgd_inrect (
    long	nlo,		/* Number of lines in output image */
    long	nrow,		/* Number of rows in mapping grid */
    long	ncol,		/* Number of columns in mapping grid */
    float	olg[],		/* Output line number on each grid row */
    float	osg[],		/* Output sample on each grid column */
    struct VTEX	v[],		/* Array of input coords and derivatives
				   at grid vertices */
    long	windopt,	/* Window specification-option code */
    long	iwindow[4],	/* Input window:  sl, ss, nl, ns */
    long	*maxnli,	/* Maximum input lines for any output
				   line */
    long	**startil,	/* Starting input line for each output
				   line */
    long	**lastil)	/* Last input line for each output
				   line */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	col;		/* Current grid column */
    float	diff;		/* Distance from separator line to
				   current grid output line value */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    long	i;		/* Loop index */
    long	icoord;		/* Coordinate on edge of input window */
    float	*il;		/* Input line coords on grid verticals */
    float	inminmax[2][2];	/* Min/max input coords on edges of
				   output image, [min[l,s],max[l,s]] */
    float	*is;		/* Input samp coords on grid verticals */
    long	*lil;		/* Last input line for each output
				   line */
    long	line;		/* Current output line or separator */
    long	missflag;	/* Set if output entire outside input
				   image */
    char	msgbuf[256];	/* Buffer for building error message */
    long	nc;		/* Number of columns in mapping grid */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    long	nr;		/* Number of rows in mapping grid */
    long	nsep;		/* Number of separator lines */
    long	outflag;	/* Set if output requires input beyond
				   edges of input image */
    long	row;		/* Current row of mapping grid */
    float	sep;		/* Output line coordinate of current
				   separator */
    float	*seplimits;	/* Minimum and maximum input line on 
				   each separator between output lines */
    float	sepminmax[2][2]; /* Min/max input coords on current
				    separator line, [min[l,s],max[l,s]] */
    long	*sil;		/* Starting input line for each output
				   line */

/*  Allocate arrays for intermediate data and final results  */

    nr = nrow;
    nc = ncol;
    nsep = nlo + 1;
    sil = *startil = (long *) malloc (2 * nlo * sizeof (long));
    lil = *lastil = sil + nlo;
    seplimits = (float *) malloc (2 * nsep * sizeof (float));
    il = (float *) malloc (2 * nc * sizeof (float));
    is = il + nc;

/*  Initialize minimum, maximum input line and pixel for entire image  */

    for (i = 0; i < 2; i++)
	{
	inminmax[0][i] = 1000000.;
	inminmax[1][i] = - 1000000.;
	}

/*  Find input minimum, maximum for each separating line  */

    row = 0;
    sep = olg[0];
    for (i = 0; i < nc; i++)
	{
	il[i] = v[i].il - v[i].dldl;
	is[i] = v[i].is - v[i].dsdl;
	}
    for (line = 0, row = 0; line < nsep; line++, sep += 1.0)
	{

/*  Check whether new grid row; if so, initialize input coords  */

	if (sep > olg[row + 1])
	    {
	    diff = sep - olg[++row] - 1.0;
	    for (col = 0; col < nc; col++)
		{
		il[col] = v[row * nc + col].il
		  + diff * v[row * nc + col].dldl;
		is[col] = v[row * nc + col].is
		  + diff * v[row * nc + col].dsdl;
		}
	    }

/*  Get input coords at each vertical grid line, and check max/min  */

	for (col = 0; col < nc; col++)
	    {
	    il[col] += v[row * nc + col].dldl;
	    is[col] += v[row * nc + col].dsdl;
	    if (col == 0)
		for (i = 0; i < 2; i++)
		    {
		    sepminmax[i][0] = il[col];
		    sepminmax[i][1] = is[col];
		    }
	    else
		rgd_ckmnmx (il[col], is[col], sepminmax);
	    }

/*  Update global max/min, and save separtor's first, last input lines  */

	for (i = 0; i < 2; i++)
	    {
	    if (sepminmax[0][i] < inminmax[0][i])
		inminmax[0][i] = sepminmax[0][i];
	    if (sepminmax[1][i] > inminmax[1][i])
		inminmax[1][i] = sepminmax[1][i];
	    seplimits[line + i * nsep] = sepminmax[i][0];
	    }
	}

/*  Determine first, last input line for each output line  */

    *maxnli = 0;
    for (line = 0; line < nlo; line++)
	{
	if (seplimits[line + 1] < seplimits[line])
	    seplimits[line] = seplimits[line + 1];
	if (seplimits[nsep + line + 1] > seplimits[nsep + line])
	    seplimits[nsep + line] = seplimits[nsep + line + 1];
	sil[line] = floor (seplimits[line] + 0.5);
	lil[line] = ceil (seplimits[nsep + line] - 0.5);
	if ((lil[line] - sil[line]) >=  *maxnli)
	    *maxnli = lil[line] - sil[line] + 1;
	}

/*  If window was specified in terms of output, adjust input window  */

    if (windopt == WOP_OUT)
	{
	outflag = FALSE;
	missflag = FALSE;
	for (i = 0; i < 2; i++)
	    {
	    icoord = floor (inminmax[0][i] + 0.5);
	    if (icoord < iwindow[i])
		outflag = TRUE;
	    else if (icoord >= (iwindow[i] + iwindow[i+2]))
		missflag = TRUE;
	    else
		{
		iwindow[i+2] -= icoord - iwindow[i];
		iwindow[i] = icoord;
		}
	    icoord = ceil (inminmax[1][i] - 0.5);
	    if ((icoord - iwindow[i]) >= iwindow[i+2])
		outflag = TRUE;
	    else if (icoord < iwindow[i])
		missflag = TRUE;
	    else
		iwindow[i+2] = icoord - iwindow[i] + 1;
	    }

/*  If needed input extends outside input image, advise user  */

	if (missflag)
	    {
	    c_errmsg ("Output window entirely outside input image",
	      "regrid-badowind", &nonfatal);
	    ABORT;
	    }
	if (outflag)
	    {
	    strcpy (msgbuf, "Output window extends outside input");
	    strcat (msgbuf, " image -- fill value will be used"); 
	    c_errmsg (msgbuf,
	      "regrid-needfill", &nonfatal);
	    }

/*  Print adjusted input window coordinates  */

	sprintf (msgbuf, "Input image window is (%ld %ld %ld %ld)",
	  iwindow[0], iwindow[1], iwindow[2], iwindow[3]);
	c_errmsg (msgbuf, "informational", &nonfatal);
	}

/*  Make all input coordinates relative to window upper left corner  */

    for (i = 0; i < nr * nc; i++)
	{
	v[i].il -= iwindow[0];
	v[i].is -= iwindow[1];
	}
    for (line = 0; line < nlo; line++)
	{
	sil[line] -= iwindow[0];
	lil[line] -= iwindow[0];
	}

/*  Return to caller  */

    free (seplimits);
    free (il);
    return;
    }			/* End of Function rgd_inrect	*/
