/*  rgd_outbox.c -- Find output window corresponding to input window

NAME:
    rgd_outbox ()

PURPOSE:
    Called by function rgd_setgrd, as part of program REGRID, to find
    the output window corresponding to a specified input window, using
    the mapping grid relating the input and output geometries.  The
    mapping grid is assumed to have been extended outwards by 0.5 output
    pixels, to define the boundary or the outer edges of the pixels
    rather than thier centers.

ARGUMENTS:
    long    iwindow[4]	(in)	Input window:  sl, ss, nl, ns.
    long    *affine	(in)	Flag set TRUE for affine transform.
    long    *nrow	(in)	Number of rows in mapping grid.
    long    *ncol	(in)	Number of columns in mapping grid.
    float   olg[]	(mod)	Output line number on each grid row.
    float   osg[]	(mod)	Output sample on each grid column.
    struct VTEX
	    v[]		(mod)	Array of input coords and derivatives
				at grid vertices.
    long    owindow[4]	(out)	Output window:  sl, ss, nl, ns.

RETURN VALUE:
    None.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in regrid.pdf.

ALGORITHM DESCRIPTION:

    Extend input window outwards 0.5 pixel, computing ulc and lrc
    Find minimum, maximum input coords on edges of extended grid
    Save input windows to undo shrinkage if affine transformation
    For each side of input window
	If side is outside mapping grid
	    Move side in to edge of mapping grid
    Determine ratio of supported area to requested input area
    If input area reduced by more than 50%
	Inform user that most of input is outside mapping grid
	If not affine
	     Abort with error message
    Else if area reduced, but by 50% or less
	Advise user that some input lines or columns are entirely
	  outside grid
    If affine transformation and input area was reduced
	Restore initial input window
    Compute coordinates of other two input corners to simplify tests
    Initialize minimum, maximum output pixels and edge/corner flags
    If not affine transformation
	For each mapping grid vertex
	    For each edge of input window
		If not last row of grid
		    If input edge cuts cell edge below grid vertex
		        Test output coords at intersection for max, min
			If first or last grid column
			    Set flag for some input outside grid
		    Else if input edge tangent to cell edge below vertex
			If grid vertex is on line between input corners
			    Test vertex coords for max, min
		If not last column of grid
		    If input edge cuts cell edge to right of grid vertex
		        Test output coords at intersection for max, min
			If first or last grid row
			    Set flag for some input outside grid
		    Else if input edge tangent to cell edge
			If grid vertex is on line between input corners
			    Test vertex coords for max, min
	    If not last vertex in row or column
		For each corner of input window
		    Compute output coords using mapping for grid
		      cell
		    If input corner is within grid cell
			Test output coords of corner for max, min
			Set flag for this input corner inside grid
    Else (affine transformation)
	For each corner of input window
	    Compute corresponding output coords
	    Test coords for max, min
    Check whether output window has been expanded
    If any input corner is outside grid or input edge cut grid boundary
      or input window adjusted previously
	Print warning that input window is partly outside mapping grid
    Compute integer coordinates of output windowo
    Return to caller
							
***************   End of Prolog for rgd_outbox   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Math library definitions		*/
#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"regrid.h"	/* Regrid-specific definitions		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    void rgd_ckmnmx (float y, float x, float minmax[2][2]);
									/*
***************   Start of Function rgd_outbox   ***************	*/

FUNCTION void rgd_outbox (
    long	iwindow[4],	/* Input window:  sl, ss, nl, ns */
    long	*affine,	/* Flag set TRUE for affine transform */
    long	*nrow,		/* Number of rows in mapping grid */
    long	*ncol,		/* Number of columns in mapping grid */
    float	olg[],		/* Output line number on each grid row */
    float	osg[],		/* Output sample on each grid column */
    struct VTEX	v[],		/* Array of input coords and derivatives
				   at grid vertices */
    long	owindow[4])	/* Output window:  sl, ss, nl, ns */
    {
	                                              			/*
Local Variable Definitions:						*/

    double	a;		/* Coefficient of quadratic eqn,
				   a*x*x + b*x + c = 0 */
    float	aratio;		/* Ratio of new to old window area */
    double	b;		/* Coefficient of quadratic eqn */
    double	c;		/* Coefficient of quadratic eqn */
    long	col;		/* Current mapping grid column */
    long	cornflag;	/* Bit flags set if corners found */
    float	denom;		/* Denominator for fraction */
    double	diff;		/* Difference between input coord at
				   adjacent grid vertices */
    double	discrim;	/* Quadratic discriminant */
    double	dist;		/* Distance from vertex to input edge */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    double	fract[2];	/* Fractional distances from vertices */
    long	gsize;		/* Total points in extended grid */
    long	i;		/* Index, loop counter */
    float	incorn[4][2];	/* Input corners, (l,s) at ul, ur, ll,
				   and lr corners */
    float	inminmax[2][2];	/* Min/max input coords on edges of
				   mapping grid, [min[l,s],max[l,s]] */
    double	intercept;	/* Other input coord. at grid
				   intersection */
    long	intercnt;	/* Number of intersections for edge */
    long	j;		/* Index, loop counter */
    long	k;		/* Index, selecting line or sample */
    long	m;		/* Loop index */
    char	msgbuf[256];	/* Buffer for building error message */
    long	nc;		/* Number of columns in mapping grid */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    long	nr;		/* Number of rows in mapping grid */
    float	oline = 0;	/* Output line value on input edge */
    float	ominmax[2][2];	/* Min/max output coords on edges of
				   mapping grid, [min[l,s],max[l,s]] */
    float	osamp;		/* Output sample value on input edge */
    long	outflag;	/* Set TRUE if input window edge cuts
				   grid boundary */
    long	rootcnt;	/* Number of roots for quadratic */
    long	row;		/* Current mapping grid row */
    float	savecorn[4][2];	/* Copy of incorn before shrinking */
    double	sqroot = 0.0;	/* Square root value from quadratic */
    float	temp;		/* Temporary storage for value */
    double	vls[4][2];	/* Input line and sample at four
				   vertices of grid cell */

/*  Extend input window outwards 0.5 pixel, computing ulc and lrc  */

    printf ("REGRID computing output window corresponding to");
    printf (" requested input window\n");
    for (i = 0; i < 2; i++)
	{
	incorn[0][i] = iwindow[i] - 0.5;
	incorn[3][i] = incorn[0][i] + iwindow[i+2];
	}

/*  Find minimum, maximum input coords on edges of extended grid  */

    nr = *nrow;
    nc = *ncol;
    gsize = nr * nc;
    for (i = 0; i < 2; i++)
	for (j = 0; j < 2; j++)
	    inminmax[i][j] = (0.5 - (float) i) * 1000000.0;
    for (col = 0; col < nc; col++)
	for (i = col; i < gsize; i += gsize - nc)
	    rgd_ckmnmx (v[i].il, v[i].is, inminmax);
    for (row = 1; row < nr - 1; row++)
        for (i = row * nc; i < (row + 1) * nc; i += nc - 1)
	    rgd_ckmnmx (v[i].il, v[i].is, inminmax);

/*  Shrink input window, if necessary, to stay within mapping grid  */

    if (*affine)
	for (i = 0; i < 4; i += 3)
	    for (j = 0; j < 2; j++)
		savecorn[i][j] = incorn[i][j];
    for (i = 0; i < 2; i++)
	{
	if (inminmax[0][i] > incorn[0][i])
	    incorn[0][i] = ceil (inminmax[0][i] + 0.5) - 0.5;
	if (inminmax[1][i] < incorn[3][i])
	    incorn[3][i] = floor (inminmax[1][i] - 0.5) + 0.5;
	}

/*  Determine ratio of supported area to requested input area  */

    aratio = ((incorn[3][0] - incorn[0][0])
      * (incorn[3][1] - incorn[0][1])) / (iwindow[2] * iwindow[3]);
    if (incorn[3][0] < incorn[0][0])
	aratio = - 1.0;

/*  Abort if shrinkage too great  */

    if (aratio < 0.5)
	{
	if (aratio <= 0.0)
	    strcpy (msgbuf, "Requested input window is entirely");
	else
	    strcpy (msgbuf, "Over half of requested input window is"); 
	    strcat (msgbuf, " outside mapping grid");
	c_errmsg (msgbuf, "regrid-badinwind", &nonfatal);
	strcpy (msgbuf, "Minimum and maximum input values in grid are");
	sprintf (msgbuf + strlen (msgbuf), " (%ld, %ld) and (%ld, %ld)",
	  (long) ceil (inminmax[0][0]), (long) ceil (inminmax[0][1]),
	  (long) floor (inminmax[1][0]), (long) floor (inminmax[1][1]));
	c_errmsg (msgbuf, "regrid-badinwind", &nonfatal);
	if (*affine)
	    {
	    sprintf (msgbuf, "%s%s", "Extending mapping grid to cover",
	      " requested window, since mapping is affine (linear)");
	    c_errmsg (msgbuf, "regrid-gridextnd", &nonfatal);
	    }
	else
	    {
	    c_errmsg ("Cannot safely extend grid", "regrid-noextend",
	      &nonfatal);
	    }
	    ABORT;
	}

/*  Merely advise user if shrinkage is less severe  */

    else if (aratio <= 0.99999)
	{
	strcpy (msgbuf, "Some lines or columns of requested input");
	strcat (msgbuf, " window are entirely outside mapping grid");
	c_errmsg (msgbuf, "regrid-inwindred", &nonfatal);
	if (! (*affine))
	    {
	    sprintf (msgbuf,
	      "Input window reduced to (%ld, %ld, %ld, %ld)",
	      (long) ceil (incorn[0][0]), (long) ceil (incorn[0][1]),
	      (long) (incorn[3][0] - incorn[0][0]),
	      (long) (incorn[3][1] - incorn[0][1]));
	    c_errmsg (msgbuf, "regrid-inwinred", &nonfatal);
	    }
	else
	    {
	    sprintf (msgbuf, "%s%s", "Extending mapping grid to cover",
	      " requested window, since mapping is affine (linear)");
	    c_errmsg (msgbuf, "regrid-gridextnd", &nonfatal);
	    }
	}

/*  If affine transformation, use full input window  */

    if ((aratio <= 0.99999) && (*affine))
	for (i = 0; i < 4; i += 3)
	    for (j = 0; j < 2; j++)
		incorn[i][j] = savecorn[i][j];

/*  Compute coordinates of other two input corners to simplify tests  */

    incorn[1][0] = incorn[0][0];
    incorn[1][1] = incorn[3][1];
    incorn[2][0] = incorn[3][0];
    incorn[2][1] = incorn[0][1];

/*  Initialize minimum, maximum output pixels and edge/corner flags  */

    for (i = 0; i < 2; i++)
	for (j = 0; j < 2; j++)
	    ominmax[i][j] = (0.5 - (float) i) * 1000000.0;
    outflag = FALSE;
    cornflag = 0;

/*  Check for input edges cutting edges of mapping grid cells  */

    if (! *affine)
	{
	for (i = 0; i < gsize; i++)
	    {
	    for (j = 0; j < 4; j++)
		{
		k = i + j + (j / 2) * (nc - 2);
		if (k < gsize)
		    {
		    vls[j][0] = v[k].il;
		    vls[j][1] = v[k].is;
		    }
		}
	    for (j = 0; j < 4; j++)
		{
		k = j % 2;
		dist = incorn[j][k] - vls[0][k];

/*  Check vertical edges of mapping grid cells first  */

		if ((i < gsize - nc)
		  && (dist * (vls[2][k] - incorn[j][k]) >= 0.0))
		    {
		    intercnt = 0;
		    diff = vls[2][k] - vls[0][k];
		    if (diff * diff > 1.0e-12)
			{
			fract[0] = dist / diff;
			intercept = vls[0][1-k]
			  + fract[0] * (vls[2][1-k] - vls[0][1-k]);
			if ((intercept >= incorn[0][1-k])
			  && (intercept <= incorn[3][1-k]))
			    {
			    intercnt = 1;
			    if ((i + 1) % nc <= 1)
				outflag = TRUE;
			    }
			}

/*  Handle special case of input edge tangent to grid vertical  */

		    else
			{
			diff = vls[2][1-k] - vls[0][1-k];
			if (diff * diff < 1.0e-12)
			    {
			    strcpy (msgbuf, "Adjacent grid vertices");
			    strcat (msgbuf, " have same input coords");
			    c_errmsg (msgbuf, "regrid-badgrid",
			      &nonfatal);
			    ABORT;
			    }
			fract[0] = (incorn[0][1-k] - vls[0][1-k]) / diff;
			fract[1] = (incorn[3][1-k] - vls[0][1-k]) / diff;
			temp = (fract[0] - 1.0) * (fract[1] - 1.0);
			if (fract[0] * fract[1] < 0.0)
			    fract[intercnt++] = 0.0;
			if ( temp < 0.0)
			    fract[intercnt++] = 1.0;
			}

/*  Compute output line number(s) at intersection point(s)  */

		    for (m = 0; m < intercnt; m++)
			{
			oline = (1.0 - fract[m]) * olg[i/nc]
			  + fract[m] * olg[(i/nc)+1];
			osamp = osg[i % nc];
			rgd_ckmnmx (oline, osamp, ominmax);
			}
		    }

/*  Now check horizontal edges of mapping grid cells  */

		if ((i % nc < nc - 1)
		  && (dist * (vls[1][k] - incorn[j][k]) >= 0.0))
		    {
		    intercnt = 0;
		    diff = vls[1][k] - vls[0][k];
		    if (diff * diff > 1.0e-12)
			{
			fract[0] = dist / diff;
			intercept = vls[0][1-k]
			  + fract[0] * (vls[1][1-k] - vls[0][1-k]);
			if ((intercept >= incorn[0][1-k])
			  && (intercept <= incorn[3][1-k]))
			    {
			    intercnt = 1;
			    if ((i < nc) || (i >= gsize - nc))
				outflag = TRUE;
			    }
			}

/*  Handle special case of input edge tangent to grid horizontal  */

		    else
			{
			diff = vls[1][1-k] - vls[0][1-k];
			if (diff * diff < 1.0e-12)
			    {
			    strcpy (msgbuf, "Adjacent grid vertices");
			    strcat (msgbuf, " have same input coords");
			    c_errmsg (msgbuf, "regrid-badgrid",
			      &nonfatal);
			    ABORT;
			    }
			fract[0] = (incorn[0][1-k] - vls[0][1-k]) / diff;
			fract[1] = (incorn[3][1-k] - vls[0][1-k]) / diff;
			temp = (fract[0] - 1.0) * (fract[1] - 1.0);
			if (fract[0] * fract[1] < 0.0)
			    fract[intercnt++] = 0.0;
			if ( temp < 0.0)
			    fract[intercnt++] = 1.0;
			}

/*  Compute output line number(s) at intersection point(s)  */

		    for (m = 0; m < intercnt; m++)
			{
			oline = olg[i / nc];
			osamp = (1.0 - fract[m]) * osg[i % nc]
			  + fract[m] * osg[(i % nc)+1];
			rgd_ckmnmx (oline, osamp, ominmax);
			}
		    }
		}

/*  Find output coords of input corners if in current grid cell  */

	    if ((i < gsize - nc) && (i % nc < nc - 1))
		{
		for (j = 0; j < 2; j++)
		    for (k =0; k < 2; k++)
			{
			inminmax[j][k] = vls[0][k];
			for (m =  1; m < 4; m++)
			    if ((inminmax[j][k] - vls[m][k])
			      * (0.5 - j) > 0.0)
				inminmax[j][k] = vls[m][k];
			}
		for (j = 0; j < 4; j++)
		    if ((incorn[j][0] >= inminmax[0][0])
		      && (incorn[j][0] <= inminmax[1][0])
		      && (incorn[j][1] >= inminmax[0][1])
		      && (incorn[j][1] <= inminmax[1][1]))

/*  Corner may be within cell -- compute its exact coordinates  */

			{
			rootcnt = 0;
			a = v[i].dldl * v[i].d2s
			  - v[i].dsdl * v[i].d2l;
			b = v[i].dldl * v[i].dsds
			  - v[i].dsdl * v[i].dlds
			  - v[i].d2s * (incorn[j][0] - v[i].il)
			  + v[i].d2l * (incorn[j][1] - v[i].is);
			c = v[i].dlds * (incorn[j][1] - v[i].is)
			  - v[i].dsds * (incorn[j][0] - v[i].il);
			if (a * a * c * c >= 1.0e-12 * b * b)
			    {
			    discrim = b * b - 4.0 * a * c;
			    if (discrim >= 0.0)
				{
				sqroot = sqrt (discrim);
				oline = (sqroot - b) / (2.0 * a);
				rootcnt = 2;
				}
			    }
			else
			    {
			    oline = - c / b;
			    rootcnt = 1;
			    }

/*  Check whether corner is in fact within cell  */

			for (k = 0; k < rootcnt; k++)
			    {
			    if ((0.0 <= oline)
			      && (oline <= olg[i/nc + 1] - olg[i/nc]))
				{
				denom = v[i].dsds
				  + v[i].d2s * oline; 
				if (denom * denom >= 1.0e-12)
				    osamp = (incorn[j][1] - v[i].is
				      - v[i].dsdl * oline) / denom;
				else
				    {
				    denom = v[i].dlds
				      + v[i].d2l * oline;
				    if (denom * denom >= 1.0e-12)
					osamp =
					  (incorn[j][1] - v[i].is 
					  - v[i].dsdl * oline)
					  / denom;
				    else
				        osamp = - 1.0;
				    }
				if ((0.0 <= osamp)
				  && (osamp <= (osg[(i % nc) + 1]
						- osg[i % nc])))
				    {
				    rgd_ckmnmx (oline + olg[i/nc],
				      osamp + osg[i % nc], ominmax);
				    cornflag |= 1<<j;
				    }
				}
			    if (rootcnt == 2)
				oline -= sqroot / a;
			    }
			}
		}
	    }

/*  If input extends outside map, advise user unless previously done  */

	if ((outflag || (cornflag != 15)) && (aratio > 0.99999))
	    {
	    strcpy (msgbuf, "Corner(s) of input window extend");
	    strcat (msgbuf, " outside mapping grid");
	    c_errmsg (msgbuf, "regrid-inwndcorn", &nonfatal);
	    strcpy (msgbuf, "Input corner(s) will not be included in");
	    strcat (msgbuf, " output image");
	    c_errmsg (msgbuf, "regrid-inunused", &nonfatal);
	    }
	}

/*  For affine transformation, simply find corners  */

    else
	{
	for (j = 0; j < 4; j++)
	    {
	    denom = v[0].dldl * v[0].dsds
	      - v[0].dlds * v[0].dsdl;
	    if (denom * denom < 1.0e-24)
		{
		c_errmsg ("Invalid mapping grid", "regrid-badaffine",
		  &nonfatal);
		ABORT;
		}
	    oline = (v[0].dsds * (incorn[j][0] - v[0].il)
	      - v[0].dlds * (incorn[j][1] - v[0].is)) / denom
	      + olg[0];
	    osamp = (v[0].dldl * (incorn[j][1] - v[0].is)
	      - v[0].dsdl * (incorn[j][0] - v[0].il)) / denom
	      + osg[0];
	    rgd_ckmnmx (oline, osamp, ominmax);
	    }
	if (((ominmax[0][0] < olg[0] || ominmax[0][1] < osg[0])
	  || (ominmax[1][0] > olg[1]) || (ominmax[1][1] > osg[1]))
	  && (aratio > 0.99999))
	    {
	    strcpy (msgbuf, "Corner(s) of input window extend");
	    strcat (msgbuf, " outside mapping grid");
	    c_errmsg (msgbuf, "regrid-inwndcorn", &nonfatal);
	    sprintf (msgbuf, "%s%s", "Extending mapping grid to cover",
	      " requested window, since mapping is affine (linear)");
	    c_errmsg (msgbuf, "regrid-gridextnd", &nonfatal);
	    }
	}

/*  Compute integer coordinates of output window  */

    for (i = 0; i < 2; i++)
	{
	owindow[i] = floor (ominmax[0][i] + 0.5);
	owindow[i+2] = ceil (ominmax[1][i] - 0.5) - owindow[i] + 1;
	}

/*  Return to caller  */

    return;
    }			/* End of Function rgd_outbox	*/
