/*  rgd_pxarea.c -- Get overlap area of input and output pixels

NAME:
    rgd_pxarea ()

PURPOSE:
    Called by rgd_tally, as part of program REGRID, to compute the
    overlap area between an input pixel and the region in input space
    covered by an output pixel.

ARGUMENTS:
    long    inls[2]	(in)	Input pixel line and sample number.
    float   ovtex[4][2] (in)	Input line and sample coordinates of
				vertices of output pixel, in clockwise
				sequence.

RETURN VALUE:
    (float) Area of overlap between input and output pixels, as fraction
	    of input pixel area.

PROGRAM HISTORY:                                  
    11-93	Initial version, R. A. White, PSU/ORSER.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Sun Sparc SunOS 4.1.3.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in regrid.pdf.

ALGORITHM DESCRIPTION:
    Compute line and sample coordinates at edges of input pixel
    Move output pixel vertices into linked list
    For each pair of adjacent output vertices
	For each edge of input pixel
	    If input edge (extended) cuts line joining output vertices
		Create new node at point of intersection
    For each output vertex
	If vertex is outside input pixel
	    If vertex is corner of output pixel
	        Remove vertex from linked list
	    Else
	        Move vertex to nearest corner of input pixel
    Initialize overlap area to 0.0
    For each triangle from first vertex to two other adjacent
      vertices 
        Compute area of triangle from cross-product of vectors
          defining two sides
        Add triangle area to overlap area
    Return total overlap area
    									
***************   End of Prolog for rgd_pxarea()   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include 	"las.h"		/* Standard LAS include files		*/
									/*
Local Structure Definitions:						*/

    struct LISTNODE	/* Node of vertex list */
	{
	struct LISTNODE *back;	/* Backward link to previous vertex	*/
	struct LISTNODE *next;  /* Forward link to next vertex */
	float ls[2];		/* Vertex line and sample coords */
	};
									/*
***************   Start of Function rgd_pxarea   ***************	*/

FUNCTION float rgd_pxarea (
    long    inls[2],		/* Input pixel line and sample number	*/
    float   ovtex[4][2])	/* Input line and sample coordinates of
				   vertices of output pixel, going
				   clockwise */
    {
	                                              			/*
Local Variable Definitions:						*/

    float	area;		/* Area of input pixel within output	*/
    struct LISTNODE
		*currvtex;	/* Pointer to current vertex in list	*/
    float	inedge[2][2];	/* Line and sample on input pixel edges	*/
    long	i;		/* Loop counter/index			*/
    long	j;		/* Loop counter/index			*/
    struct LISTNODE
		*listhead;	/* Pointer to first active vertex	*/
    long	nextfree;	/* Index of next free vertex entry	*/
    struct LISTNODE
		vertex[20];	/* Linked list of overlap area vertices	*/


/*  Compute line and sample coordinates at edges of input pixel  */

    for (i = 0; i < 2; i++)
        for (j = 0; j < 2; j++)
	    inedge[i][j] = (float) inls[i] - 0.5 + (float) j;

/*  Move output pixel vertices into linked list  */

    listhead = &vertex[0];
    for (nextfree = 0; nextfree < 4; nextfree++)
	{
	vertex[nextfree].back = &(vertex[(nextfree + 3) % 4]);
	vertex[nextfree].next = &vertex[(nextfree + 1) % 4];
	for (i = 0; i < 2; i++)
	    {
	    vertex[nextfree].ls[i] = ovtex[nextfree][i];
	    }
	}

/*  Examine (possibly truncated) output edges for input crossings  */

    for (currvtex = listhead; currvtex != NULL;
      currvtex = (currvtex->next != listhead) ? currvtex->next : NULL)
	{
	for (i = 0; i < 2; i++)
	    {
	    for (j = 0; j < 2; j++)
		{
		if (((currvtex->ls[i] < inedge[i][j])
		  && (currvtex->next->ls[i] > inedge[i][j]))
		  || ((currvtex->ls[i] > inedge[i][j])
		  && (currvtex->next->ls[i] < inedge[i][j])))

/*  Create new node at point of intersection  */

		    {
		    vertex[nextfree].ls[i] = inedge[i][j];
		    vertex[nextfree].ls[1-i] = currvtex->ls[1-i]
		      + (currvtex->next->ls[1-i] - currvtex->ls[1-i])
			* (inedge[i][j] - currvtex->ls[i])
			/ (currvtex->next->ls[i] - currvtex->ls[i]);
		    vertex[nextfree].next = currvtex->next;
		    vertex[nextfree].back = currvtex;
		    currvtex->next->back = &vertex[nextfree];
		    currvtex->next = &vertex[nextfree++];
		    }
		}
	    }
	}

/*  Delete original output vertices outside input pixel  */

    for (i = 0; i < 4; i++)
	{
	if ((vertex[i].ls[0] < inedge[0][0])
	  || (vertex[i].ls[0] > inedge[0][1])
	  || (vertex[i].ls[1] < inedge[1][0])
	  || (vertex[i].ls[1] > inedge[1][1]))
	    {
	    vertex[i].back->next = vertex[i].next;
	    vertex[i].next->back = vertex[i].back;
	    if (&vertex[i] == listhead)
	        listhead = vertex[i].next;
	    }
	}

/*  Move other vertices outside input pixel to nearest corner  */

    for (currvtex = listhead ; currvtex != NULL;
      currvtex = (currvtex->next != listhead) ? currvtex->next : NULL)
	{
	if (currvtex->ls[0] < inedge[0][0])
	    currvtex->ls[0] = inedge[0][0];
	else if (currvtex->ls[0] > inedge[0][1])
	    currvtex->ls[0] = inedge[0][1];
	else if (currvtex->ls[1] < inedge[1][0])
	    currvtex->ls[1] = inedge[1][0];
	else if (currvtex->ls[1] > inedge[1][1])
	    currvtex->ls[1] = inedge[1][1];
	}

/*  Compute area of overlap -- area of triangle defined by successive
    vertices is given by cross product of vectors defining edges   */

    area = 0.0;
    for (currvtex = listhead->next; currvtex->next != listhead;
      currvtex = currvtex->next)
	{
	area += 0.5 * ((currvtex->ls[1] - listhead->ls[1])
		* (currvtex->next->ls[0] - listhead->ls[0])
		- (currvtex->ls[0] - listhead->ls[0])
		* (currvtex->next->ls[1] - listhead->ls[1]));
	}

/*  Return overlap area  */

    return area;
    }			/* End of Function rgd_pxarea */
