/*  RGD_SETGRD.C -- Set up mapping grid and compute gradients

NAME:
    rgd_setgrd ()

PURPOSE:
    Called by REGRID to get the mapping grid, compute gradients within
    each grid cell, adjust the input window if necessary, and trim or
    extend the grid to just contain the output window.

ARGUMENTS:
    char    *gridfile	(in)	Name of mapping grid file.
    long    windopt	(in)	Window option code.
    long    iwindow[4]	(in)	Input window specification.
    long    owindow[4]	(mod)	Output window specification.
    struct GEOGRID
	    *grid	(out)	Mapping grid, unmodified.
    long    *nrow	(out)	Number of rows in mapping grid, after
				adjustment 
    long    *ncol	(out)	Number of columns in adjusted grid.
    float   **outline	(out)	Output line number on each grid row.
    float   **outsamp	(out)	Output sample on each grid column.
    struct VTEX
	    **vertex	(out)	Input line and sample values at grid
				vertices and derivatives with respect
				to output coords.

RETURN VALUE:
    None.

FILE REFERENCES:
    gridfile	File containing mapping grid relating output geometry to
		input geometry.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Sun Sparc SunOS 4.1.3 and Solaris 2.3.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in REGRID.PDF.

EXTERNAL FUNCTIONS CALLED:
    rgd_outbox	Determine output window corresponding to requested input
		window.
    rgd_trmgrd	Trim or expand mapping grid to coincide with specified
		output window.

ALGORITHM DESCRIPTION:
    Get mapping grid <c_getgrd>
    Make local copies of grid data
    Compute derivatives of input coordinates relative to output
    Determine whether transformation is affine
    If WINDOPT is IN or output window defaulted
	Set output window from mapping grid
    Else
	If requested output window extends beyond mapping grid
	    Inform user
	    If not affine transformation
		Abort processing
    Adjust grid to be 0.5 pixel larger than output window <rgd_trmgrd>
    If WINDOPT is IN
	Determine output box needed to enclose input window <rgd_outbox>
	If output window not same as grid coverage
	    Adjust grid again to match new output window <rgd_trmgrd>
    Make output coordinates relative to output window

***************   End of Prolog for rgd_setgrd   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String processing functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"geompak.h"	/* Definition of mapping grid structure */
#include	"regrid.h"	/* Regrid-specific definitions		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    void rgd_outbox (long iwindow[4], long *affine, long *nrow,
		     long *ncol, float olg[], float osg[],
		     struct VTEX v[], long owindow[4]);
    void rgd_trmgrd (float outedges[4], long *nrow, long *ncol,
		     float olg[], float	osg[], struct VTEX v[]);
									/*
***************   Start of Function rgd_setgrd   ***************	*/

FUNCTION void rgd_setgrd (
    char	*gridfile,	/* Name of mapping grid file */
    long	windopt,	/* Window option code */
    long	iwindow[4],	/* Input window specification */
    long	owindow[4],	/* Output window specification */
    struct GEOGRID
		*grid,		/* Mapping grid, unmodified */
    long	*nrow,		/* Number of rows in mapping grid,
				   after adjustment */
    long	*ncol,		/* Number of columns in adjusted grid */
    float	**outline,	/* Output line number on each grid row */
    float	**outsamp,	/* Output sample on each grid column */
    struct VTEX	**vertex)	/* Input line and sample values at grid
				   vertices and derivatives with respect
				   to output coords */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	affine;		/* Flag set TRUE if affine transform */
    long	col;		/* Current grid column */
    float	denom;		/* Denominator for gradient */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    long	gsize;		/* Total vertices in mapping grid */
    long	i;		/* Loop index */
    long	modflag;	/* Flag set true if owindow modified */
    char	msgbuf[256];	/* Buffer for building error message */
    long	nc;		/* Local copy of number of grid cols */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    long	nr;		/* Local copy of number of rows */	
    float	*olg;		/* Output line values for each row of
				   grid, in floating point */
    float	*osg;		/* Output sample values for each row of
				   grid, in floating point */
    float	outedges[4];	/* Edges of extended output window */
    long	row;		/* Current grid row */
    long	saveow[4];	/* Copy of unadjusted output window */
    struct VTEX	*v;		/* Array of input coords and derivatives
				   at grid vertices */

/*  Get mapping grid  */

    if (c_getgrd (gridfile, grid) != E_SUCC)
	ABORT;

/*  Make local copies of mapping grid entries  */

    nr = *nrow = grid->nrows;
    nc = *ncol = grid->ncols;
    olg = *outline = (float *) malloc (nr * sizeof(long));
    osg = *outsamp = (float*) malloc (nc * sizeof(long));
    for (row = 0; row < nr; row++)
	olg[row] = grid->out_lines[row];
    for (col = 0; col < nc; col++)
	osg[col] = grid->out_samps[col];
    gsize = nr * nc;
    v = *vertex = (struct VTEX *) malloc (8 * gsize * sizeof(*v));
    for (i = 0; i < gsize; i++)
	{
	v[i].il = grid->in_lines[i];
	v[i].is = grid->in_samps[i];
	}

/*  Compute derivatives of input coords relative to output  */

    for (row = 0, i = 0; row < nr; row++)
        for (col = 0; col < nc; col++, i++)
	    {
	    if (row < nr - 1)
		{
		denom = (olg[row + 1] - olg[row]);
		v[i].dldl = (v[i + nc].il - v[i].il) / denom;
		v[i].dsdl = (v[i + nc].is - v[i].is) / denom;
		if (col < nc - 1)
		    {
		    denom = ((olg[row + 1] - olg[row])
		      * (osg[col + 1] - osg[col])); 
		    v[i].d2l = (v[i + nc].il + v[i + 1].il
		      - v[i].il  - v[i + nc + 1].il) / denom;
		    v[i].d2s = (v[i + nc].is + v[i + 1].is
		      - v[i].is  - v[i + nc + 1].is) / denom;
		    }
		}
	    if (col < nc - 1)
		{
		denom = (osg[col + 1] - osg[col]);
		v[i].dlds = (v[i + 1].il - v[i].il) / denom;
		v[i].dsds = (v[i + 1].is - v[i].is) / denom;
		}
	    }

/*  Determine whether transformation is affine  */

    if ((nr == 2) && (nc == 2) && (v[0].d2l < 0.01 * iwindow[2])
      && (v[0].d2s < 0.01 * iwindow[3]))
	affine = TRUE;
    else
	affine = FALSE;		 

/*  Establish initial output window  */

    if (windopt == WOP_IN)
	{
	owindow[0] = owindow[1] = 1;
	owindow[2] = olg[nr-1] - 1;
	owindow[3] = osg[nc-1] - 1;
	}
    else if (owindow[2] < 1)
	{
	owindow[0] = owindow[1] = 1;
	owindow[2] = grid->lines;
	owindow[3] = grid->samples;
	}

/*  Check whether user requested output window outside grid area  */

    else if ((owindow[0] < 1) || (owindow[1] < 1)
      || (owindow[2] > grid->out_lines[nr-1])
      || (owindow[3] > grid->out_samps[nc-1]))
	{

	strcpy (msgbuf, "Requested output window extends outside area");
	sprintf (msgbuf + strlen (msgbuf),
	   " defined by mapping grid (%d, %d, %ld, %ld) -- ", 1, 1,
	   grid->out_lines[nr-1], grid->out_samps[nc-1]);
	c_errmsg (msgbuf, "regrid-bigowind", &nonfatal);
	if (affine)
	    {
	    sprintf (msgbuf, "%s%s", "Extending mapping grid to cover",
	      "requested window, since mapping is affine");
	    c_errmsg (msgbuf, "regrid-gridextend", &nonfatal);
	    }
	else
	    {
	    c_errmsg ("Cannot safely extend grid", "regrid-noextend",
	      &nonfatal);
	    ABORT;
	    }
	}

/*  Adjust grid to 0.5 greater than output window on all sides  */

    for (i = 0; i < 2; i++)
	{
	outedges[i] = owindow[i] - 0.5;
	outedges[i+2] = outedges[i] + owindow[i+2];
	}
    rgd_trmgrd (outedges, nrow, ncol, olg, osg, v);

/*  If input window specified, determine corresponding output window  */

    if (windopt == WOP_IN)
	{
	for (i = 0; i < 4; i++)
	    saveow[i] = owindow[i];
	rgd_outbox (iwindow, &affine, nrow, ncol, olg, osg, v, owindow);

/*  If output window changed, adjust grid to fit  */

	for (i = 0, modflag = FALSE; i < 4; i++)
	    if (saveow[i] != owindow[i])
		modflag = TRUE;
	if (modflag)
	    {
	    for (i = 0; i < 2; i++)
		{
		outedges[i] = owindow[i] - 0.5;
		outedges[i+2] = outedges[i] + owindow[i+2];
		}
	    rgd_trmgrd (outedges, nrow, ncol, olg, osg, v);
	    }
	}

/*  Make output coordinates relative to output window  */

    for (i = 0; i < *nrow; i++)
	olg[i] -= owindow[0];
    for (i = 0; i < *ncol; i++)
	osg[i] -= owindow[1];

/*  Return to caller  */

    return;
    }			/*  End of function rgd_setgrd  */
