/*  rdg_tally.c -- Tally each input pixel's contribution to output value

NAME:
    rgd_tally ()

PURPOSE:
    Called by rgd_aggreg as part of program REGRID to tally the
    contributions of each input pixel to the current output pixel and
    set the output pixel value as specified by the program argument
    FUNCTION.

ARGUMENTS:
    long    aggrcode	(in)	Aggregation function code.
    long    bndndx[]	(in)	Index into inbuf for start of each band.
    float   fill	(in)	Fill value for pixels outside input.
    long    idcode	(in)	Input image datatype.
    long    idsize	(in)	Bytes per input pixel.
    unsigned char
	    *inbuf	(in)	Input image buffer.
    long    maxnli	(in)	Maximum input lines for any output.
    long    nbndin	(in)	Number of input bands.
    long    nbndout	(in)	Number of bands in output image.
    long    nclass	(in)	Number of classes.
    long    nli		(in)	Number of lines in input window.
    long    nsi		(in)	Number of samples in input window.
    long    ntclass	(in)	Number of classes with thresholds.
    long    nwclass	(in)	Number of classes with weights != 1.
    long    odcode	(in)	LAS data type code of output image.
    long    odsize	(in)	Bytes per output pixel.
    long    offsets[]	(in)	Output buffer offsets for successive
				output bands.
    float   ovtex[4][2]	(in)	Input line and sample coordinates of
				vertices of output pixel.
    float   scalefact	(in)	Scaling factor for FRACT output.
    struct WTR
	    *wtr	(in)	Class weights, thresholds, and
				reallocation specifications.
    float   accum[]	(mod)	Accumulators for area-weighted totals
				(preallocated scratch space).
    unsigned char
	    *outbuf	(mod)   Output buffer for first output band.

RETURN VALUE:
    None.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    1994-07	Added class reallocation, R. A. White, PSU/ESSC.
    2001-08	Added NODATA test for subcommand MODE, raw.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Sun Sparc SunOS 4.1.3.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in REGRID.PDF.

ALGORITHM DESCRIPTION:

    (Note: for output pixel straddling grid cell boundary, any
     kinks in the input mapping at boundary can be ignored, since
     grid cell size was chosen to cause mapping errors much less
     than one pixel; for same reason, there is no possibility of
     square output pixel mapping into concave polygon in input.)

    Determine smallest rectangular array of input pixels which wholly
      contains output pixel
    Clear accumulators for weighted values
    For each input pixel in rectangle
	If input pixel not entirely outside input image
	    if only one pixel in rectangle
		Area = 1.0
	    Else
		Compute overlap area of input, output pixel <rgd_pxarea>
	    Add area to total area
	    If FUNCTION is FRACT or MODE with one input band
		Add area to accumlator specified by input pixel value
	    Else  (MODE with multiple input bands or FUNCTION = MEAN)
		For each input band
		    If NODATA specified and equal to value
			Add area to nodata accumulator for band
		    Else
			Multiply pixel value for band by area and add to 
			  accumlator for band
    If class reallocation was specified and FUNCTION = MODE
	For each class to be reallocated
	    Save total area for class and clear class accumulator
	    For each new class to receive reallocation
		Mulitply old class area by reallocation weight and add
		  to total for new class
	    Adjust total area to reflect sum of new weights
    If weighting values were specified and FUNCTION = MODE
	For each weight
	    Adjust total area to reflect weighted class area
	    Multiply accumulator for specified class by weight value
    If total area is 0.0
	Enter fill value into all output bands
    Else
	For each accumlator
	    If NODATA specified
		If nodata area equals total area
		    Enter fill value for output band
		Else
		    Divide accumulator by non-NODATA area
	    Else
		Divide contents of accumulator by total area
	If FUNCTION is MODE
	    If thresholds are specified
		For each class having value at or below threshold
		    Negate fractional value
	    Find maximum value for any class
	    If maximum is zero
		Find class with most negative value
	Else
	    Copy accumulator values to respective output bands
    									
***************   End of Prolog for rgd_tally   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Math library definitions		*/
#include	<stdio.h>	/* Standard I/O definitions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"regrid.h"	/* Definitions unique to regrid		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    float rgd_pxarea (long inls[2], float ovtex[4][2]);
									/*
***************   Start of Function rgd_tally   ***************	*/

void FUNCTION rgd_tally (
    long	aggrcode,	/* Aggregation function code */
    long	bndndx[],	/* Index into inbuf for start of each
				   band */
    float	fill,		/* Fill value for pixels outside input */
    long	idcode,		/* Input image datatype */
    long	idsize,		/* Bytes per input pixel */
    unsigned char
		*inbuf,		/* Input image buffer */
    long	maxnli,		/* Maximum input lines for any output */
    long	nbndin,		/* Number of input bands */
    long	nbndout,	/* Number of bands in output image */
    long	nclass,		/* Number of classes */
    long	nli,		/* Number of lines in input window */
    long	nsi,		/* Number of samples in input window */
    long	odcode,		/* LAS data type code of output image */
    long	odsize,		/* Bytes per output pixel */
    long	offsets[],	/* Output buffer offsets for successive
				   output bands */
    float	ovtex[4][2],	/* Input line and sample coordinates of
				   vertices of output pixel */
    float	scalefact,	/* Scaling factor for FRACT output */
    struct WTR	*wtr,		/* Class weights, thresholds, and
				   reallocation specifications */
    float	accum[],	/* Accumulators for area-weighted
				   totals */
    unsigned char
		*outbuf)	/* Output buffer for first output band */
    {
	                                              			/*
Local Variable Definitions:						*/

    float	area;		/* Overlap area for one input pixel */
    float	*fractptr;	/* Next reallocation fraction entry */
    long	i;		/* Loop counter/index */
    unsigned char
		*ibptr;		/* Location of current input pixel in
				   input buffer */
    long	il;		/* Line number of input pixel */
    long	inbox[2][2];	/* Integer version of inminmax */
    long	inls[2];	/* Line and sample coord of input pixel */
    float	inminmax[2][2];	/* Minimum and maximum input line and
				   sample needed for output pixel */
    long	is;		/* Sample number of input pixel */
    long	j;		/* Loop counter/index */
    long	maxclass;	/* Class having maximum weighted area */
    long	maxrclass = -1;	/* Class having maxrval */
    float	maxval;		/* Maximum weighted area of any class */
    float	maxrval;	/* Negative of maximum value rejected in
				   threshold test */
    long	*newclptr;	/* Next reallocation new class entry */
    unsigned char
		*obptr;		/* Location of current output band in
				   output buffer */
    float	savearea;	/* Area of class being reallocated */
    float	totarea;	/* Total overlap area with input */
    float	totweight;	/* Total reallocation weight */
    double	value = 0.0;	/* Value of input pixel */

/*  Find smallest input rectangle enclosing output pixel  */

    for (i = 0; i < 2; i++)
	{
	inminmax[0][i] = inminmax[1][i] = ovtex[0][i];
	for (j = 1; j < 4; j++)
	    {
	    if (inminmax[0][i] > ovtex[j][i])
		inminmax[0][i] = ovtex[j][i];
	    if (inminmax[1][i] < ovtex[j][i])
		inminmax[1][i] = ovtex[j][i];
	    }
	inbox[0][i] = floor (inminmax[0][i] + 0.5);
	if (inbox[0][i] < 0)
	    inbox[0][i] = 0;
	inbox[1][i] = ceil (inminmax[1][i] - 0.5);
	}
    if (inbox[1][0] >= nli)
	inbox[1][0] = nli - 1;
    if (inbox[1][1] >= nsi)
	inbox[1][1] = nsi - 1;

/*  Clear accumulator for each class  */

    for (i = 0; i < nclass; i++)
	{
	accum[i] = 0.0;
	if (wtr->ndflag > 0)
	    accum[i + nclass] = 0.0;
	}
    totarea = 0.0;

/*  Find overlap area of each input pixel with output pixel  */

    for (il = inbox[0][0]; il <= inbox[1][0]; il++)
	{
	ibptr = inbuf  + idsize * (nsi * (il % maxnli) + inbox[0][1]);
	for (is = inbox[0][1]; is <= inbox[1][1];
	  is++, ibptr += idsize)
	    {
	    if ((inbox[0][0] == inbox[1][0])
	      && (inbox[0][1] == inbox[1][1]))
		area = 1.0;
	    else
		{
		inls[0] = il;
		inls[1] = is;
		area = rgd_pxarea (inls, ovtex);
		}

/*  Add area to appropriate accumulators  */

	    totarea += area;
	    if (((aggrcode == RGD_MODE) && (nbndin == 1))
	      || (aggrcode == RGD_FRACT))
		{
		switch (idcode)
		    {
		    case EBYTE:
		    case ECHAR:
			i = *ibptr;
			break;
		    case EWORD:
			i = * ( (short*) ibptr);
			break;
		    case ELONG:
			i = * ( (long*) ibptr);
			break;
		    case EREAL:
			i = * ( (float*) ibptr);
			break;
		    case EDOUBLE:
			i = * ( (double*) ibptr);
		    }
		accum[i] += area;
		}

/*  Multiply area by input value for FUNCTION = MEAN or multiband MODE	*/

	    else
		for (i = 0; i < nbndin; i++)
		    {
		    switch (idcode)
			{
			case EBYTE:
			case ECHAR:
			    value = *(ibptr + bndndx[i]);
			    break;
			case EWORD:
			    value =
			      * ( (short*) (ibptr + bndndx[i]));
			    break;
			case ELONG:
			    value =
			      * ( (long*) (ibptr + bndndx[i]));
			    break;
			case EREAL:
			    value =
			      * ( (float*) (ibptr + bndndx[i]));
			    break;
			case EDOUBLE:
			    value =
			      * ( (double*) (ibptr + bndndx[i])); 
			}

/*  Check for NODATA value  */

		    if ((wtr->ndflag > 0) && (value == wtr->nodata))
			accum[i + nclass] += area;
		    else
			accum[i] += area * value;
		    }
	    }
	}

/*  Reallocate class areas, if requested  */

    if (wtr->rcount > 0)
	{
	newclptr = wtr->newcl;
	fractptr = wtr->fract;
	for (i = 0; i < wtr->rcount; i++)
	    {
	    totweight = 0.0;
	    savearea = accum[wtr->rclass[i]];
	    accum[wtr->rclass[i]] = 0.0;
	    for (j = 0; j < wtr->nclcnt[i]; j++)
		{
		accum[*newclptr++] += savearea * *fractptr;
		totweight += *fractptr++;
		}
	    totarea += savearea * (totweight - 1.0);
	    }
	}

/*  Multiply accumulators by weights, if any  */

    if (wtr->wcount > 0)
	for (i = 0; i < wtr->wcount; i++)
	    {
	    j = wtr->weight[i];
	    totarea += accum[j] * (wtr->weight[i] - 1.0);
	    accum[j] *= wtr->weight[i];
	    }

/*  If total area is zero, use fill value  */

    if (totarea <= 0.0)
	{
	for (i = 0; i < nclass; i++)
	    accum[i] = fill;
	}

/*  Convert accumulator values to relative areas  */

    else
	{
	for (i = 0; i < nclass; i++)
	    {
	    if (wtr->ndflag > 0)
		{
		if (accum[i + nclass] >= totarea)
		    accum[i] = fill;
		else
		    accum[i] /= (totarea - accum[i + nclass]);
		}
	    else
		accum[i] /= totarea;
	    }

/*  Eliminate any class which doesn't reach threshold for mode	*/

	if (aggrcode == RGD_MODE)
	    {
	    if (wtr->tcount > 0)
		{
		maxrval = 0.0;
		for (i = 0; i < wtr->tcount; i++)
		    {
		    j = wtr->tclass[i];
		    if (accum[j] < wtr->thresh[i])
			{
			accum[j] *= -1.0;
			if (accum[j] < maxrval)
			    {
			    maxrval = accum[j];
			    maxrclass = j;
			    }
			}
		    }
		}

/*  Find mode  */

	    maxval = 0.0;
	    maxclass = -1;
	    for (i = 0; i < nclass; i++)
		if (accum[i] > maxval)
		    {
		    maxval = accum[i];
		    maxclass = i;
		    }
	    if (maxclass < 0)
		maxclass = maxrclass;
	    accum[0] = maxclass;
	    }
	}

/*  Place results in output buffer  */

    for (i = 0; i < nbndout; i++)
	{
	obptr = outbuf + odsize * offsets[i];
	if (aggrcode == RGD_FRACT)
	    accum[i] = accum[i+1] * scalefact;
	switch (odcode)
	    {
	    case EBYTE:
	    case ECHAR:
	        *obptr = (unsigned char) (accum[i] + 0.5);
	        break;
	    case EWORD:
	        * ( (short *) obptr) = (short) (accum[i] + 0.5);
	        break;
	    case ELONG:
	        * ( (long *) obptr) = (long) (accum[i] + 0.5);
	        break;
	    case EREAL:
	        * ( (float *) obptr) = accum[i];
	        break;
	    case EDOUBLE:
	        * ( (double *) obptr) = accum[i];
	    }
	}

   return;
   }		/* End of Function rgd_tally */
