/*  RGD_TRMGRD.C -- Trim mapping grid to specified output window

NAME:
    rgd_trmgrd ()

PURPOSE:
    Called by rgd_setgrd () as part of program REGRID to trim a mapping
    grid to just match a specified (not necessarily integer) output
    window.  If rows or columns of the grid are eliminated, the grid is
    repacked to squeeze them out.

ARGUMENTS:
    float    outedges[4]   (in)    Output window edges of trimmed grid,
				   (l, s) at ulc, (l, s) at lrc.
    long     *nrow	   (mod)   Number of rows in mapping grid.
    long     *ncol	   (mod)   Number of columns in mapping grid.
    float    olg[]	   (mod)   Output line number on each grid row.
    float    osg[]	   (mod)   Output sample on each grid column.
    struct VTEX
	     v[]	   (mod)   Array of input coords and derivatives
				   at grid vertices.

RETURN VALUE:
    None.

PROGRAM HISTORY:                                  
    1993-12	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Sun Sparc SunOS 4.1.3.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in REGRID.PDF.

ALGORITHM DESCRIPTION:
    Determine first and last row and column of grid needed
    If grid size is reduced
	Repack grid to squeeze out unused cells
    For each cell in top and bottom row
	Compute input coordinates and derivatives at requested first and
	  last output line
    For each cell in first and last column
	Compute input coordinates and derivatives at requested first and
	  last output sample
    Return to caller
    									
***************   End of Prolog for rgd_trmgrd ()   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"regrid.h"	/* Definitions unique to regrid		*/
									/*
***************   Start of Function rgd_trmgrd   ***************	*/

void FUNCTION rgd_trmgrd (
    float	outedges[4],	/* Output window edges of trimmed grid,
				   (l, s) at ulc, (l, s) at lrc */
    long	*nrow,		/* Number of rows in mapping grid */
    long	*ncol,		/* Number of columns in mapping grid */
    float	olg[],		/* Output line number on each grid row */
    float	osg[],		/* Output sample on each grid column */
    struct VTEX	v[])		/* Array of input coords and derivatives
				   at grid vertices */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	col;		/* Current mapping grid column */
    float	delta1;		/* Distance from top/right of grid to
				   requested ulc coordinate */
    float	delta2;		/* Distance from bottom/left of grid to
				   requested lrc coordinate */
    long	fatal = -1;	/* Fatal error flag for c_errmsg */
    long	gsize;		/* Total points in extended grid */
    long	i;		/* Index, loop counter */
    long	ic;		/* Initial column of mapping grid needed
				   for output window */
    long	ir;		/* Initial row of mapping grid needed
				   for output window */
    long	lc;		/* Last column of mapping grid needed
				   for output window */
    long	lr;		/* Last row of mapping grid needed for
				   output window */
    long	nc;		/* Number of grid columns */
    long	nonfatal = 2;	/* Warning flag for c_errmsg */
    long	nr;		/* Number of grid rows */
    long	row;		/* Current mapping grid row */

/*  Determine first and last row and column of grid needed  */

    if ((outedges[0] >= outedges[2]) || (outedges[1] >= outedges[3]))
	{
	c_errmsg ("Output window has negative dimension(s)",
	  "regrid-badowind", &nonfatal);
	ABORT;
	}
    nr = *nrow;
    nc = *ncol;
    for (ir = 0; (ir < nr - 2) && (outedges[0] >= olg[ir+1]); ir++);
    for (ic = 0; (ic < nc - 2) && (outedges[1] >= osg[ic+1]); ic++);
    for (lr = nr - 1; (lr > 1) && (outedges[2] <= olg[lr-1]); lr--);
    for (lc = nc - 1; (lc > 1) && (outedges[3] <= osg[lc-1]); lc--);

/*  Squeeze out unused grid cells, if any  */

    if (((lr - ir) < (nr - 1)) || ((lc - ic) < (nc - 1)))
	{
	for (i = 0, row = ir; row <= lr; row++)
	    for (col = ic; col <= lc; col++, i++)
		v[i] = v[row * nc + col];
	for (i = 0, row = ir; row <= lr; row++, i++)
	    olg[i] = olg[row];
	for (i = 0, col = ic; col <= lc; col++, i++)
	    osg[i] = osg[col];
	*nrow = nr = lr - ir + 1;
	*ncol = nc = lc - ic + 1;
	}

/*  Adjust input coordinates and derivatives on top and bottom row  */

    delta1 = outedges[0] - olg[0];
    delta2 = outedges[2] - olg[nr-1];
    gsize = nr * nc;
    for (col = 0, i = gsize - nc; col < nc; col++, i++)
	{
	v[col].il += delta1 * v[col].dldl;
	v[col].is += delta1 * v[col].dsdl;
	if (col != nc - 1)
	    {
	    v[col].dlds += delta1 * v[col].d2l;
	    v[col].dsds += delta1 * v[col].d2s;
	    }
	v[i].il += delta2 * v[i-nc].dldl;
	v[i].is += delta2 * v[i-nc].dsdl;
	}
    olg[0] += delta1;
    olg[nr-1] += delta2;

/*  Adjust input coordinates and derivatives on right and left column  */

    delta1 = outedges[1] - osg[0];
    delta2 = outedges[3] - osg[nc-1];
    for (row = 0, i = nc - 1; row < gsize; row += nc, i += nc)
	{
	v[row].il += delta1 * v[row].dlds;
	v[row].is += delta1 * v[row].dsds;
	if (row != nc - 1)
	    {
	    v[row].dldl += delta1 * v[row].d2l;
	    v[row].dsdl += delta1 * v[row].d2s;
	    }
	v[i].il += delta2 * v[i-1].dlds;
	v[i].is += delta2 * v[i-1].dsds;
	}
    osg[0] += delta1;
    osg[nc-1] += delta2;

/*  Return to caller  */

    return;
    }			/* End of Function rgd_trmgrd	*/
